#' @title Sequencing To Imputation Through Constructing Haplotypes
#' @param chr What chromosome to run. Should match BAM headers
#' @param posfile Where to find file with positions to run. File is tab seperated with no header, one row per SNP, with col 1 = chromosome, col 2 = physical position (sorted from smallest to largest), col 3 = reference base, col 4 = alternate base. Bases are capitalized. Example first row: 1<tab>1000<tab>A<tab>G<tab>
#' @param K How many founder / mosaic haplotypes to use
#' @param Number of generations since founding or mixing. Note that the algorithm is relatively robust to this. Use nGen = 4 * Ne / K if unsure
#' @param outputdir What output directory to use
#' @param tempdir What directory to use as temporary directory. If possible, use ramdisk, like /dev/shm/
#' @param bamlist Path to file with bam file locations. File is one row per entry, path to bam files. Bam index files should exist in same directory as for each bam, suffixed either .bam.bai or .bai
#' @param cramlist Path to file with cram file locations. File is one row per entry, path to cram files. cram files are converted to bam files on the fly for parsing into STITCH
#' @param reference Path to reference fasta used for making cram files. Only required if cramlist is defined
#' @param genfile Path to gen file with high coverage results. Empty for no genfile. File has a header row with a name for each sample, matching what is found in the bam file. Each subject is then a tab seperated column, with 0 = hom ref, 1 = het, 2 = hom alt and NA indicating missing genotype, with rows corresponding to rows of the posfile. Note therefore this file has one more row than posfile which has no header
#' @param method How to run imputation - either diploid or pseudoHaploid, the former being the original method quadratic in K, the later being linear in K
#' @param outputInputInVCFFormat Whether to output the input in vcf format
#' @param downsampleToCov What coverage to downsample individual sites to. This ensures no floating point errors at sites with really high coverage
#' @param downsampleFraction Downsample BAMs by choosing a fraction of reads to retain. Must be value 0<downsampleFraction<1
#' @param readAware Whether to run the algorithm is read aware mode. If false, then reads are split into new reads, one per SNP per read
#' @param chrStart When loading from BAM, some start position, before SNPs occur. Default NA will infer this from either regionStart, regionEnd and buffer, or posfile
#' @param chrEnd When loading from BAM, some end position, after SNPs occur. Default NA will infer this from either regionStart, regionEnd and buffer, or posfile
#' @param regionStart When running imputation, where to start from. The 1-based position x is kept if regionStart <= x <= regionEnd
#' @param regionEnd When running imputation, where to stop
#' @param buffer Buffer of region to perform imputation over. So imputation is run form regionStart-buffer to regionEnd+buffer, and reported for regionStart to regionEnd, including the bases of regionStart and regionEnd
#' @param maxDifferenceBetweenReads How much of a difference to allow the reads to make in the forward backward probability calculation. For example, if P(read | state 1)=1 and P(read | state 2)=1e-6, re-scale so that their ratio is this value. This helps prevent any individual read as having too much of an influence on state changes, helping prevent against influence by false positive SNPs
#' @param alphaMatThreshold Minimum (maximum is 1 minus this) state switching into probabilities
#' @param emissionThreshold Emission probability bounds. emissionThreshold < P(alt read | state k) < (1-emissionThreshold)
#' @param iSizeUpperLimit Do not use reads with an insert size of more than this value
#' @param bqFilter Minimum BQ for a SNP in a read. Also, the algorithm uses bq<=mq, so if mapping quality is less than this, the read isnt used
#' @param niterations Number of EM iterations
#' @param shuffleHaplotypeIterations Iterations on which to perform heuristic attempt to shuffle founder haplotypes for better fit. To disable set to NA.
#' @param splitReadIterations Iterations to try and split reads which may span recombination breakpoints for a better fit
#' @param nCores How many cores to use
#' @param expRate Expected recombination rate in cM/Mb
#' @param maxRate Maximum recomb rate cM/Mb
#' @param minRate Minimum recomb rate cM/Mb
#' @param Jmax Maximum number of SNPs on a read
#' @param regenerateInput Whether to regenerate input files
#' @param originalRegionName If regenerateInput is FALSE (i.e. using existing data), this is the name of the original region name (chr.regionStart.regionEnd). This is necessary to load past variables
#' @param keepInterimFiles Whether to keep interim parameter estimates
#' @param keepTempDir Whether to keep files in temporary directory
#' @param environment Whether to use server or cluster multicore options
#' @param pseudoHaploidModel How to model read probabilities in pseudo diploid model (shouldn't be changed)
#' @param diploidModel How to model read probabilities in diploid model (shouldn't be changed)
#' @param outputHaplotypeProbabilities Whether to output haplotype probabilities in files
#' @param switchModelIteration Whether to switch from pseudoHaploid to diploid and at what iteration (NA for no switching)
#' @param generateInputOnly Whether to just generate input data then quit
#' @param restartIterations In pseudoHaploid method, which iterations to look for collapsed haplotype prnobabilities to resolve
#' @param refillIterations When to try and refill some of the less frequently used haplotypes
#' @param downsampleSamples What fraction of samples to retain. Useful for checking effect of N on imputation. Not meant for general use
#' @param downsampleSamplesKeepList When downsampling samples, specify a numeric list of samples to keep
#' @param subsetSNPsfile If input data has already been made for a region, then subset down to a new set of SNPs, as given by this file. Not meant for general use
#' @param useSoftClippedBases Whether to use (TRUE) or not use (FALSE) bases in soft clipped portions of reads
#' @param outputBlockSize How many samples to write out to disk at the same time when making temporary VCFs that are later pasted together at the end to make the final VCF. Smaller means lower RAM footprint, larger means faster write. 
#' @return Results in properly formatted version
#' @author Robert Davies
#' @export
STITCH <- function(
  chr,
  nGen="",
  posfile,
  K,
  outputdir,
  tempdir,
  bamlist = "",
  cramlist = "",
  reference = "",
  genfile = "",
  method = "diploid",
  outputInputInVCFFormat = FALSE,
  downsampleToCov = 50,
  downsampleFraction = 1,
  readAware = TRUE,
  chrStart = NA,
  chrEnd = NA,
  regionStart = NA,
  regionEnd = NA,
  buffer = NA,
  maxDifferenceBetweenReads = 1000,
  alphaMatThreshold = 1e-4,
  emissionThreshold = 1e-4,
  iSizeUpperLimit = as.integer(600),
  bqFilter = as.integer(17),
  niterations = 40,
  shuffleHaplotypeIterations = c(4,8,12,16),
  splitReadIterations = 25,
  nCores = 1,
  expRate = 0.5,
  maxRate = 100,
  minRate = 0.1,
  Jmax = 100,
  regenerateInput = TRUE,
  originalRegionName = NA,
  keepInterimFiles = FALSE,
  keepTempDir = FALSE,
  environment = "server",
  pseudoHaploidModel = 9,
  diploidModel = 2,
  outputHaplotypeProbabilities = FALSE,
  switchModelIteration = NA,
  generateInputOnly = FALSE,
  restartIterations = NA,
  refillIterations = c(6,10,14,18),
  downsampleSamples = 1,
  downsampleSamplesKeepList = NA,
  subsetSNPsfile = NA,
  useSoftClippedBases = FALSE,
  outputBlockSize = 1000
) {


  #
  # specify output description
  #
  regionName=chr
  options(scipen=999) # Dangit rounding 
  if(is.na(regionStart)==FALSE & is.na(regionEnd)==FALSE)
    regionName=paste(chr,".",regionStart,".",regionEnd,sep="")


  
  
  #
  # define some other variables - remove eventually
  #
  highCovInLow=NULL # note - why is this null?
  outputImpute2=FALSE # whether to output references in impute2 format
  outputInputGen=FALSE # whether to output input in gen format
  startIterations=0 # how many start iterations to run
  windowSNPs=5000 # if doing the window start, how many SNPs to loop over
  if(method=="diploid") restartIterations=NA  
  #
  #
  # run some checks
  #
  #
  if(chr=="") stop("missing chr - need to specify chromosome")
  if(nGen=="") stop("need to set nGen")
  if(posfile=="") stop("missing pos - need to specify sites to impute over")
  if(K=="") stop("missing K")
  if(outputdir=="") stop("missing outputdir")
  if(tempdir=="") stop("missing tempdir")
  if(outputBlockSize<1) stop("outputBlockSize must be an integer greater than 0")
  if(round(outputBlockSize)!=outputBlockSize) stop("outputBlockSize must be an integer greater than 0")
  #
  # extra checks on region start, end, buffer
  #
  w <- as.integer(is.na(regionStart))+
       as.integer(is.na(regionEnd))+
       as.integer(is.na(buffer))
  if ( w !=0 & w!=3)
    stop("either all three of regionStart, regionEnd and buffer are NA, or none of them are")
  if (is.na(regionStart)==FALSE) {
     if (regionStart<1) stop("regionStart must be >0")
     if (round(regionStart)!=regionStart) stop("regionEnd must be an integer")
  }
  if (is.na(regionEnd)==FALSE) {
     if (regionEnd<1) stop("regionEnd must be >0")
     if (round(regionEnd)!=regionEnd) stop("regionEnd must be an integer")
  }
  if (is.na(buffer)==FALSE) {
     if (buffer<1) stop("buffer must be >0")
     if (round(buffer)!=buffer) stop("buffer must be an integer")
  }
  if (is.na(regionStart)==FALSE & is.na(regionEnd)==FALSE) {
    if (regionStart>regionEnd) stop("regionStart must be before regionEnd")
  }
  if (regenerateInput == FALSE) {
    if (is.na(originalRegionName) == TRUE) stop("if regenerateInput is FALSE (i.e. using existing data), you must supply the original region name (must not be NA) to load the input properly. Also don't forget to supply the position file used to make the original input data")
    if (bamlist != "" || cramlist != "") stop("If not regenerating input, please do not supply bamlist")
    if (is.na(regionStart) | is.na(regionEnd) | is.na(buffer)) stop("If regenerateInput is FALSE, please supply original regionStart, regionEnd and buffer used to make the inputs")
  }
  #
  # extra checks on ability to use CRAM files
  #
  if (cramlist != "") {
    if (bamlist != "")
      stop("Please specify only one or cramlist or bamlist")
    if (reference == "")
      stop("If you supply a list of CRAM files, you need to supply the reference they were aligned against")
    system(paste0("which samtools > ", tempdir(), "A.txt 2>&1"))
    line <- readLines(paste0(tempdir(), "A.txt"))
    if (strsplit(line, ": ")[[1]] == "which") {
      stop("Cannot find samtools in path. Please add samtools to path using export PATH=${PATH}:/path/to/samtools/bin")
    }
  }
  #
  #
  # make necessary directory
  #
  #
  tempdir=paste(tempdir,paste(toupper(letters[sample(26,10,replace=TRUE)]),collapse=""),"/",sep="")
  system(paste("mkdir -p ",tempdir,sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  system(paste("mkdir -p ",outputdir,sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  system(paste("mkdir ",outputdir,"RData/",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  system(paste("mkdir ",outputdir,"debug/",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)  
  system(paste("mkdir ",outputdir,"plots/",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  inputdir=paste(outputdir,"input/",sep="")
  system(paste("mkdir ",inputdir,sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  if(outputImpute2==TRUE) system(paste("mkdir ",outputdir,"impute2format/",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)


  #
  # print out date
  #
  date=date()
  print(paste(c("Program start - ",date),collapse=""))
  if(generateInputOnly==FALSE)
    save(date,file  =paste(outputdir,"RData/start.",regionName,".RData",sep=""))
  
  #
  # save parameters
  #
  # save.image(paste(outputdir,"RData/imputation.parameters.",chr,".RData",sep="") )
  #
  #
  # load the positions (and high coverage genotypes)
  #
  #
  pos=read.table(posfile,header=FALSE,sep="\t")
  # check chromosomes agree
  if(sum(pos[,1]!=chr)>0) stop("ERROR - pos file not unique to chromosome - check column 1")
  if(sum(diff(order(pos[,2]))!=1)>0) stop("ERROR - pos file not sorted on position")
  if(sum(nchar(as.character(pos[,3]))!=1)>0) stop("ERROR - pos file multiple reference alleles")
  if(sum(nchar(as.character(pos[,4]))!=1)>0) stop("ERROR - pos file multiple alternate alleles")
  colnames(pos)=c("CHR","POS","REF","ALT")
  L=as.integer(as.character(pos[,"POS"]))
  T=as.integer(length(L))
  # load high coverage genotype files if available
  gen=NULL
  if(genfile!="")
  {
    gen=read.table(genfile,header=TRUE,sep="\t")
    # argh R screwing up some filenames
    colnames(gen)=as.character(unlist(read.table(genfile,sep="\t",nrows=1)))
    if(nrow(gen)!=nrow(pos)) stop("ERROR - gen and pos must be the same size (other than the header in pos")
  }


  #
  # shrink
  #
  if(is.na(regionStart)==FALSE & is.na(regionEnd)==FALSE)
  {
    # need to shrink pos, L, gen
    # then, need to fix input
    inRegion=L>=(regionStart-buffer) & L<=(regionEnd+buffer)
    inRegionL=range((1:length(inRegion))[inRegion]) # this may be used later on
    L=L[inRegion]
    pos=pos[inRegion,]
    gen=gen[inRegion,]
    T=as.integer(nrow(pos))
  } else {
    inRegionL <- c(1, T)
  }


  #
  # determine chrStart and chrEnd
  #
  if ((as.integer(is.na(chrStart)) + as.integer(is.na(chrEnd))) == 1)
    stop("Please set either both or neither of chrStart and chrEnd to NA")
  if (is.na(chrStart) & is.na(chrEnd)) {
    chrStart <- min(L) - iSizeUpperLimit
    chrEnd <- max(L) + iSizeUpperLimit
  }
  
  
  #
  #
  # get sample names
  #
  #
  bam_files <- NULL
  cram_files <- NULL
  if(bamlist!="" | cramlist!="")
  {
    if (cramlist != "") {
      cram_files <- as.character(read.table(cramlist)[,1])      
      sampleNames <- get_cram_sample_names(cram_files, nCores)
    } else if (bamlist != "") {
      bam_files <- as.character(read.table(bamlist)[,1])
      fileHeaders=scanBamHeader(bam_files) # surprisingly slow!
      x=lapply(fileHeaders,function(x) x$text[["@RG"]])
      y=lapply(x,function(z) z[grep("SM",z)])
      sampleNames=as.character(unlist(lapply(y,function(a) substr(a,4,100))))
    }
    if(length(unique(sampleNames))!=length(sampleNames)) stop("ERROR - repeat sample names") 
    # save sample names
    save(sampleNames, file=paste(outputdir,"RData/sampleNames.",regionName,".RData",sep=""))
  } else {
    # otherwise, we won't be re-doing vinput anyway!
    # load up sampleNames, fileHeaders
    file <- paste0(outputdir,"RData/sampleNames.", originalRegionName,  ".RData")
    if (file.exists(file) == TRUE) {
      load(file)
    } else {
      stop(paste0("Cannot find ",file))
    }
  }
  # 
  N <- as.integer(length(sampleNames)) # number of samples
  #
  # load in sequencing data
  #
  flag=     scanBamFlag(isPaired = NA, isProperPair = NA, isUnmappedQuery = FALSE, hasUnmappedMate = NA, isMinusStrand = NA, isMateMinusStrand = NA, isFirstMateRead = NA, isSecondMateRead = NA, isNotPrimaryRead = NA, isNotPassingQualityControls = FALSE, isDuplicate = FALSE)
  ### okay actually start loading here
  what=c("qname","strand","pos","seq","qual","cigar","isize","mapq")
  ### set which region of the genome to interrogate
  eval(parse(text=  (paste("which = RangesList(\"",chr,"\"=IRanges(",chrStart,",",chrEnd,"))",sep=""))))
  #
  # if the data doesn't exist, rebuild it
  #
  if(regenerateInput==TRUE) {
    print(paste(c("Generate inputs - ",date()),collapse=""))
    if(environment=="server")
    {
      out=mclapply(as.list(1:N),FUN=loadBamAndConvert,which=which,what=what,L=L,pos=pos,T=T,bam_files = bam_files,cram_files = cram_files,reference=reference,iSizeUpperLimit=iSizeUpperLimit,bqFilter=bqFilter,chr=chr,outputdir=outputdir,N=N,mc.cores=nCores,downsampleToCov=downsampleToCov,sampleNames=sampleNames,flag=flag,inputdir=inputdir,useSoftClippedBases=useSoftClippedBases, regionName = regionName, tempdir = tempdir, chrStart = chrStart, chrEnd = chrEnd)
    }
    if(environment=="cluster")
    {
      cl = makeCluster(nCores, type = "FORK")
      out = parLapply(cl, as.list(1:N), fun=loadBamAndConvert,which=which,what=what,L=L,pos=pos,T=T,bam_files = bam_files,cram_files = cram_files,reference=reference,iSizeUpperLimit=iSizeUpperLimit,bqFilter=bqFilter,chr=chr,outputdir=outputdir,N=N,mc.cores=nCores,downsampleToCov=downsampleToCov,sampleNames=sampleNames,flag=flag,inputdir=inputdir,useSoftClippedBases=useSoftClippedBases, regionName = regionName, tempdir = tempdir, chrStart = chrStart, chrEnd = chrEnd)
      stopCluster(cl)
    }
    if(generateInputOnly==TRUE) {
      # save pos file for next time
      save(pos,file=paste(outputdir,"RData/pos.",regionName,".RData",sep=""))
      print(paste0("Done generating input - ",date()))
      return(NULL)
    }
  }
  
  #
  # some more things
  #
  ### figure out where high coverage are in low coverage data
  nHC=0
  if(genfile!="")
  {
    ### set highCovList from this
    highCovList=colnames(gen)
    nHC=length(highCovList)
    #=c("Q_CFW-SW___145.0h","Q_CFW-SW___145.0b","Q_CFW-SW___143.0p","Q_CFW-SW___47.0p")
    highCovInLow=array(0,nHC)
    for(i in 1:nHC) highCovInLow[i]=  match(highCovList[i],sampleNames)
    if(sum(is.na(highCovInLow))>0) stop("ERROR - high coverage sample not found in low coverage samples")
  }


  

  

  #
  #
  # at this point, decide if need to shrink BAMs (only if regenerateInput = FALSE)
  #
  #
  if(regenerateInput == TRUE) {
    # ie we've just remade the data
    # can copy over directly now
    system(paste("cp ",inputdir,"sample*.input.", regionName, ".RData ",tempdir,sep=""))
  } else {
    if (regionName == originalRegionName) {
    system(paste("cp ",inputdir,"sample*.input.", regionName, ".RData ",tempdir,sep=""))
    } else {
      # other-wise - not rebuilding
      # shrink the BAMs anyway - safer - in case the buffer has changed
      print(paste(c("Shrink inputs - ",date()),collapse=""))    
      if(environment=="server")
      {
        out=mclapply(as.list(1:N),FUN=shrinkReads,mc.cores=nCores,inputdir=inputdir,regionName=regionName, originalRegionName = originalRegionName,chr=chr,tempdir=tempdir,inRegionL=inRegionL,savedir=tempdir)
      }
      if(environment=="cluster")
      {
        cl = makeCluster(nCores, type = "FORK")
        out2 = parLapply(cl, as.list(1:N),fun=shrinkReads,inputdir=inputdir,regionName=regionName, originalRegionName = originalRegionName,chr=chr,tempdir=tempdir,inRegionL=inRegionL,savedir=tempdir)
        stopCluster(cl)
      }
    }
  }


  

  #
  # if we want, re-intersect with new set of positions
  #
  if(is.na(subsetSNPsfile)==FALSE & subsetSNPsfile!="NA")
  {
    print(paste("Subsetting SNPs from file ",subsetSNPsfile,sep=""))
    # load in positions
    out=subsetSNPsFunction(N=N,subsetSNPsfile=subsetSNPsfile,regionName=regionName,tempdir=tempdir,L=L,environment=environment,nCores=nCores,outputdir=outputdir)
    print(paste("Back to main ",subsetSNPsfile,sep=""))    
    keep=out$keep
    T=as.integer(sum(keep))
    L=L[keep]
    pos=pos[keep,]
    gen=gen[keep,]
    print(paste("Done subsetting section ",subsetSNPsfile,sep=""))
    save(T,L,keep,pos,gen,file=paste(outputdir,"debug/files.RData",sep=""))
    print(T)
  }

  


  #
  # downsample the number of samples
  #
  # (note - mostly for testing - not for general use)
  if(downsampleSamples<1)
  {
    keepList=read.table(downsampleSamplesKeepList)
    keep=array(FALSE,N)
    # number of samples
    keep[unlist(keepList)]=TRUE
    # now - sample some more
    keep[keep==FALSE][sample(1:sum(keep==FALSE),N*downsampleSamples-length(unlist(keepList)))]=TRUE
    #
    # now - we know which we want to keep
    # keep them, and re-name appropriately
    #
    N=sum(keep)
    keepL=(1:length(keep))[keep]
    for(iSample in 1:N) # new N
    {
      new=iSample
      old=keepL[iSample]
      # argh - move to temp, them move back
      system(paste("mv ",tempdir,"sample.",old,".input.",regionName,".RData ",tempdir,"sample.","XXX",".input.",regionName,".RData",sep=""))
      system(paste("mv ",tempdir,"sample.","XXX",".input.",regionName,".RData ",tempdir,"sample.",new,".input.",regionName,".RData",sep=""))
    } # done
    save(keepL,file=paste(outputdir,"RData/keepL.",regionName,".RData",sep=""))
  }
  



  
  #
  # downsample to percent
  #
  if(downsampleFraction!=1)
    downsampleToFraction(N=N,nCores=nCores,downsampleFraction=downsampleFraction,regionName=regionName,tempdir=tempdir,environment=environment)
  #
  # build alleleCount
  #
  print("Generate allele count")  
  alleleCount=buildAlleleCount(T=T,N=N,nCores=nCores,regionName=regionName,tempdir=tempdir,environment=environment)
  alleleCount[is.na(alleleCount[,3]),3]=0 # shouldn't be in here...
  print("Done generating allele count")    
  #
  # toggle read aware
  #
  if(readAware==FALSE)
  {
    print("Split reads")
    splitReadsCompletely(N=N,nCores=nCores,tempdir=tempdir,regionName=regionName,environment=environment)
    print("Done splitting reads")
  }



  #
  # if we're so inclined, output input in VCF format
  #
  if (outputInputInVCFFormat == TRUE) {
    out=outputInputInVCFFunction(outputdir = outputdir, pos = pos, T = T, tempdir = tempdir, N = N, nCores = nCores, regionName = regionName, environment = environment, sampleNames = sampleNames, outputBlockSize = outputBlockSize)
    return(NULL)
  }
    

  
  
  #
  #
  #
  #
  # initialize variables
  #
  #
  #
  a1=0; b1=1
  eHapsCurrent=matrix(a1+(b1-a1)*runif(T*K),ncol=K)
  if(startIterations>0) # but, make the rest of them uniform - get info later
    eHapsCurrent[(2*windowSNPs+1):T,]=0.25
  # initialize using rate assuming 0.5 cM/Mb, T=100
  sigmaCurrent=exp(-nGen*expRate/100/1000000*diff(L)) # interior is morgans between SNPs
  # alphaMatList - where to switch
  alphaMatCurrent=matrix(1/K,nrow=(T-1),ncol=K)
  # prior list
  priorCurrent=rep(1/K,K)
  hapSumCurrent=array(1/K,c(T,K))



  


  
  #
  #
  # run EM algorithm here
  # 
  #
  #
  print(paste("Begin EM - ",date()))
  date=date()
  save(date,file  =paste(outputdir,"RData/startEM.",regionName,".RData",sep=""))
  print(paste("Number of SNPs - ",T,sep=""))
  ### actually run now
  iteration=max(0,startIterations)
  #iteration=39;load(file=paste(outputdir,"RData/interim.",regionName,".iteration",iteration,".RData",sep=""))
  #iteration=0
  #while(iteration<12)
  while(iteration<niterations)
  {
    iteration=iteration+1
    #
    # switch methods if appropriate
    #
    if(is.na(switchModelIteration)==FALSE) {
      if(iteration==switchModelIteration) {
        method="diploid"
        print(paste("Switching from pseudoHaploid to diploid - iteration ",iteration," - ",date(),sep=""))
    } }
    #
    # fork out and get results
    #
    p=mcparallel(completeSampleIteration(N=N,tempdir=tempdir,chr=chr,K=K,T=T,priorCurrent=priorCurrent,eHapsCurrent=eHapsCurrent,alphaMatCurrent=alphaMatCurrent,sigmaCurrent=sigmaCurrent,maxDifferenceBetweenReads=maxDifferenceBetweenReads,whatToReturn=whatToReturn,Jmax=Jmax,highCovInLow=highCovInLow,iteration=iteration,method=method,expRate=expRate,minRate=minRate,maxRate=maxRate,niterations=niterations,splitReadIterations=splitReadIterations,shuffleHaplotypeIterations=shuffleHaplotypeIterations,nCores=nCores,L=L,nGen=nGen,emissionThreshold=emissionThreshold,alphaMatThreshold=alphaMatThreshold,nHC=nHC,gen=gen,outputdir=outputdir,environment=environment,pseudoHaploidModel=pseudoHaploidModel,diploidModel=diploidModel,outputHaplotypeProbabilities=outputHaplotypeProbabilities,switchModelIteration=switchModelIteration,regionName=regionName,restartIterations=restartIterations,refillIterations=refillIterations,hapSumCurrent=hapSumCurrent,outputBlockSize=outputBlockSize))
    c=mccollect(p)
    #
    #
    #
    if(length(c[[1]])==1) # error - stop
    {
      print(iteration)
      iteration=niterations+1
      save(c,file=paste(outputdir,"RData/mainDebug.",regionName,".RData",sep=""))
      save(eHapsCurrent,sigmaCurrent,alphaMatCurrent,priorCurrent,file=paste(outputdir,"RData/interim.chr",regionName,".iteration",iteration,".RData",sep=""))
      save.image(file=paste(outputdir,"RData/image.",regionName,".iteration",iteration,".RData",sep=""))
      stop(paste("ERROR in iterations - see file",outputdir,"RData/mainDebug.",regionName,".RData",sep=""))
    } else {
      # save previous results to disk
      eHapsFuture=c[[1]]$eHapsFuture
      alphaMatFuture=c[[1]]$alphaMatFuture
      sigmaFuture=c[[1]]$sigmaFuture
      priorFuture=c[[1]]$priorFuture
      hapSumFuture=c[[1]]$hapSum
      # new - get hwe, info, estimatedAlleleFrequency
      hwe=c[[1]]$hwe
      info=c[[1]]$info
      estimatedAlleleFrequency=c[[1]]$estimatedAlleleFrequency
      # save everything if needed
      save(hapSumFuture,eHapsCurrent,alphaMatCurrent,sigmaCurrent,priorCurrent,eHapsFuture,alphaMatFuture,sigmaFuture,priorFuture,file=paste(outputdir,"RData/interim.",regionName,".iteration",iteration,".RData",sep=""))
      # perform switchover here
      eHapsCurrent=eHapsFuture
      alphaMatCurrent=alphaMatFuture
      sigmaCurrent=sigmaFuture
      priorCurrent=priorFuture
      hapSumCurrent=hapSumFuture
      hapSum=hapSumCurrent
      plotHapSum(outname=paste(outputdir,"plots/hapSum.",regionName,".iteration.",iteration,".jpeg",sep=""),L=L,which=1:T,K=K,hapSum=hapSum,T=T,N=N)
    }
  }


  #
  # build final VCF from pieces
  #
  write_vcf_after_EM(outputdir = outputdir, regionName = regionName, sampleNames = sampleNames, tempdir = tempdir, nCores = nCores, info = info, hwe = hwe, estimatedAlleleFrequency = estimatedAlleleFrequency, pos = pos, N = N, outputBlockSize = outputBlockSize)
  print(paste("Done EM and outputting - ",date()))
  

  #
  # get imputed dosages back in matrix form
  #
  if(nHC > 0) {
    gen_imp <- get_high_coverage_final_imputed_genotypes(highCovInLow = highCovInLow, nHC = nHC, tempdir = tempdir, regionName = regionName, T = T)
  }

  
  #
  #
  # now, if there was a buffer, remove the unwanted SNPs
  #
  #
  if(is.na(regionStart)==FALSE & is.na(regionEnd)==FALSE)
  {
    print(paste0("Remove buffer region from output - ",date()))
    # determine where the region is
    inRegion2=L>=regionStart & L<=regionEnd
    inRegion2L=(1:length(inRegion2))[inRegion2]
    ### pos, gen, L
    pos=pos[inRegion2,]
    gen=gen[inRegion2,]
    if ( nHC > 0 )
      gen_imp <- gen_imp[inRegion2, ]
    alleleCount=alleleCount[inRegion2, ]
    L=L[inRegion2]
    T=as.integer(nrow(pos))
    ### parameters and hapSum
    x=inRegion2L; x=x[-length(x)]
    alphaMatCurrent=alphaMatCurrent[x,]
    sigmaCurrent=sigmaCurrent[x]    
    eHapsCurrent=eHapsCurrent[inRegion2,]
    hapSumCurrent=hapSumCurrent[inRegion2,]
    priorCurrent=hapSum[1,]/sum(hapSum[1,])
    ###
    info <- info[inRegion2]
    hwe <- hwe[inRegion2]    
    estimatedAlleleFrequency <- estimatedAlleleFrequency[inRegion2]    
    #
    # shrink the VCF
    #
    print(paste0("Remove buffer from VCF - ",date()))
    file <- paste0(outputdir,"stitch.",regionName,".vcf.gz")
    file_temp <- paste0(outputdir,"stitch.",regionName,".temp.vcf.gz")
    command <- paste0("zless ",file," | cat | awk '{if(substr($0, 1, 1)==",'"#"'," || ( $2>=", regionStart, " && $2<=", regionEnd, ")) {print $0}}' | gzip -1 > ",file_temp)
    system(command)
    system(paste0("mv ", file_temp, " ", file))
  } # end of cleaning up the VCF and other objects


  
    





  #
  # save various important objects to disk
  #
  print(paste0("Save RData objects to disk - ",date()))  
  save(info, file=paste(outputdir,"RData/info.",regionName,".RData",sep=""))
  save(hwe, file=paste(outputdir,"RData/hwe.",regionName,".RData",sep=""))
  passQC <- info > 0.4 & hwe > 1e-6 # one interpretation of passQC
  save(passQC, file=paste(outputdir,"RData/passQC.",regionName,".RData",sep=""))
  save(hapSum, file=paste(outputdir,"RData/hapSum.",regionName,".RData",sep=""))


  
  #
  # save important files from EM output (ie eHaps, niterations)
  #
  # save final parameter values
  save(eHapsCurrent,alphaMatCurrent,sigmaCurrent,priorCurrent,hapSumCurrent,file=paste(outputdir,"RData/EMfinalestimates.",regionName,".RData",sep=""))
  save(alleleCount,estimatedAlleleFrequency,pos,gen,L,file=paste(outputdir,"RData/EMnecessary.",regionName,".RData",sep=""))
  # save particulars
  save(nCores,N,T,chr,K,niterations,file=paste(outputdir,"RData/EMparameters.",regionName,".RData",sep=""))
  if(nHC > 0)
    save(gen_imp,file=paste(outputdir,"RData/gen_imp.",regionName,".RData",sep=""))
  


  



  
  #
  #
  # do some plotting
  #
  #
  print(paste0("Make plots - ",date()))
  plotMetricsForPostImputationQC(iSample=NULL,highCovList=NULL,gen=gen, gen_imp = gen_imp, alleleCount=alleleCount,chr=chr,L=L,estimatedAlleleFrequency=estimatedAlleleFrequency,info=info,outputdir=outputdir,hwe=hwe,regionName=regionName)
  #
  # plot hapProbs sum - should tell what states are being used
  #
  plotHapSum(outname=paste(outputdir,"plots/hapProbs.run.",regionName,".jpg",sep=""),L=L,which=1:T,K=K,hapSum=hapSum,T=T,N=N)
  #
  ### plot estimated AF against real 0.1X pileups (get r2 as well)
  #
  if(sum(passQC)>1)
    plotEstimatedAgainstReal(outputdir=outputdir,alleleCount=alleleCount,estimatedAlleleFrequency=estimatedAlleleFrequency,which=passQC,chr=chr,regionName=regionName)






  #
  # clean up and terminate
  #
  print(paste0("Clean up and end - ",date()))  
  warnings()
  # remove things from tempdir
  if(keepTempDir==FALSE) 
  {  
    system(paste("rm ",tempdir,"sample.","*",".input.",regionName,".RData",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
    system(paste("rm ",tempdir,"sample*RData",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
    system(paste("rmdir ",tempdir,sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
  }
  if(keepInterimFiles==FALSE)
    system(paste("rm ",outputdir,"RData/interim.",regionName,"*RData",sep=""),ignore.stdout=TRUE,ignore.stderr=TRUE)
    

  date=date()
  print(paste(c("Program done - ",date),collapse=""))
  save(date, file = paste(outputdir,"RData/end.",regionName,".RData",sep=""))

  # done
  return(NULL)
  
}








# for an iteration, find intervals of where there are infrequently
# used haplotypes
getWhereToRefill=function(hapSum,T,hapFreqMin)
{
  # 0.5% frequency for a hap
  hapFreqMin=0.005
  # find runs of a certain length less than this value
  whereToRefill=lapply(1:K,hapSum=hapSum,T=T,hapFreqMin=hapFreqMin,function(k,hapFreqMin,T,hapSum) {
    x=hapSum[,k]
    # those less than the minimum
    y=as.integer(x<hapFreqMin*N)
    # find break points
    z=(y[-1]-y[-length(y)])!=0
    start=c(1,(1:(T-1))[z]+1)
    end=c((1:(T-1))[z],T)
    m=matrix(cbind(start,end,y[start]),ncol=3)
    colnames(m)=c("start","end","insufficient")
    # only keep where we need to redo
    m=matrix(m[m[,3]==1,]    ,ncol=3)
    # now - remove some of them - keep if >100
    if(nrow(m)==1)      return(m)
    #
    # remove those smaller than 100 
    #
    m=matrix(m[(m[,2]-m[,1])>100,],ncol=3)
    return(m)
  })
  return(whereToRefill)
}





get_cram_sample_names <- function(
  cram_files,
  nCores
) {
  
  print(paste0("Get CRAM sample names - ", date()))
  sampleNames <- mclapply(
    cram_files,
    mc.cores = nCores,
    function(cram_file) {
      header <- system(paste0("samtools view -H ", cram_file, " | grep @RG"), intern = TRUE)
      #header <- header[substr(header, 1, 3) == "@RG"]
      rg_spots <- lapply(header, strsplit, split = "\t")
      if (length(rg_spots) == 0)
        stop(paste0("There is no @RG tag (with sample name) for:", cram_file))
      sm <- sapply(rg_spots, function(x) {
        rg <- x[[1]]
        sm <- substr(rg[substr(rg, 1, 3) == "SM:"], 4, 1000)
        return(sm)
      })
      if (length(unique(sm)) > 1)
        stop(paste0("There is more than one sample name in the header for:", cram_file))
      return(sm[1])
    }
  )
  
  sampleNames <- as.character(unlist(sampleNames))
  print(paste0("Done getting  CRAM sample names - ", date()))  
  return(sampleNames)
}



# in a simple refill, look every 100 SNP. in each one with an average below a threshold, sample from other haplotypes to re-copy
refillSimple=function(hapSum,T,K,eHapsCurrent,N)
{
  hapFreqMin=0.005 # percentage
  nSNP=100
  r=lapply(1:floor(T/nSNP),function(x) c(1+nSNP*(x-1),nSNP*x)) #
  a=t(sapply(r,function(x) {
    colSums(hapSum[x[1]:x[2],])/nSNP
  }))
  b=a<(N*hapFreqMin)
  # for each k, for each region, fill in
  # within each continuous region, fill in with respect to frequencies of all other haplotypes
  #
  for(k in 1:K)
  {
    # change into intervals
    z1=b[,k]
    x=(1:(length(z1)-1))[diff(z1)!=0]
    start=c(1,x+1)
    end=c(x,length(z1))
    # only take "bad" regions
    start=start[z1[start]==TRUE]
    end=end[z1[end]==TRUE    ]
    if(length(start)>0)
    {
      for(iR in 1:length(start))
      {
        # now - in each region, get sums of haplotypes
        p=colSums(matrix(a[start[iR]:end[iR],],ncol=K)); p=p/sum(p)
        replacement=sample(K,1,prob=p)
        y=r[[start[iR]]][1]:r[[end[iR]]][2]
        eHapsCurrent[y,k]=eHapsCurrent[y,replacement]
      }
    }
  } # end! done replacement
  print(paste("Refill infrequently used haplotypes - ",round(100*sum(b==TRUE)/(nrow(b)*ncol(b)),1),"% of regions replaced",sep=""))
  return(list(eHapsCurrent=eHapsCurrent))
  #for(i in 1:length(r))
  #{
  #  good=b[i,]==FALSE
  #  goodV=(1:K)[good]
  #  bad=b[i,]==TRUE
  #  badV=(1:K)[bad]
  #  # get frequencies of other haps
  #  u=a[i,good];     u=u/sum(u)
  #  # sample from these frequencies
  #  toReplace=sample(x=goodV,size=sum(good==FALSE),prob=u,replace=TRUE)
  #  # take common haps, fill in
  #  y=r[[i]][1]:r[[i]][2]
  #  for(iR in 1:length(toReplace))
  #    eHapsCurrent[y,badV[iR]]=eHapsCurrent[y,toReplace[iR]]
  #}
}
















loadBamAndConvert=function(iBam,which,what,L,pos,T,bam_files,cram_files,reference,iSizeUpperLimit,bqFilter,chr,outputdir,N,downsampleToCov,sampleNames,flag,inputdir,useSoftClippedBases, regionName, tempdir, chrStart, chrEnd)
{
  ###
  ###
  ###
  if (length(bam_files) > 0) {
    bamName <- bam_files[iBam]
  } else if (length(cram_files) > 0) {
    cramName <- cram_files[iBam]
    tempBam <- paste0(tempdir, "/sample", iBam, ".bam")
    # note - chrStart and chrEnd are appropriate
    system(paste0(
      "samtools view -h ", cramName,
      " -T ", reference, " ",
      chr, ":", chrStart, "-", chrEnd, " | ",
      "samtools view -b > ",
      tempBam
    ))
    system(paste0("samtools index ", tempBam))
    bamName <- tempBam
  } else {
    stop("Both bam_files and cram_files are empty")
  }
  if((iBam%%100)==0)
    print(paste("load and convert bam ",iBam," of ",N,", " ,date(),sep=""))
  ### load in a single result
  ###
  #idx=paste(substr(bamName,1,nchar(bamName)-3),"bam.bai",sep="")
  idx1=paste(substr(bamName,1,nchar(bamName)-3),"bai",sep="")
  idx2=paste(bamName,".bai",sep="")
  if(file.exists(idx1)) idx=idx1
  if(file.exists(idx2)) idx=idx2
  param=ScanBamParam(flag=flag,which=which,what=what) # define parameters
  sampleData=scanBam(file=bamName,index=idx,param=param) # load the data
  if (length(cram_files) > 1)
    system(paste0("rm ", tempdir, "sample", iBam, ".ba*"))
  #
  # first, a quick check on whether there are reads
  #
  if(length(sampleData[[1]]$qname)==0)
  {
    # no reads on this chromosome! output a warning and bogus data
    print(paste("WARNING - sample ",sampleNames[iBam]," has no reads. It is being given random reads. Consider removing from analysis",sep=""))
    sampleReads=lapply(1:5,function(i) list(0,matrix(as.integer(i),ncol=1,nrow=1),matrix(as.integer(i),ncol=1,nrow=1),matrix(as.integer(i),ncol=1,nrow=1)))      
        # nU,d,phiU,pRU
        # 1st - nU = 0-based number in read
        # 2nd - d = "representative" position, 0-based index
        # 3rd - phiU = for each SNP in read, probability of alt allele
        # 4th - pRU = for each SNP in read, 0-based index in L
  } else {
    ###
    ### get info on reads
    ###
    # get more info read as well
    mapq=as.integer(sampleData[[1]]$mapq)
    isize=sampleData[[1]]$isize
    # figure out which ones we dont want - remove
    keep=FALSE==(mapq<bqFilter | ( abs(isize)>iSizeUpperLimit & is.na(isize)==FALSE))
    cigarRead=sampleData[[1]]$cigar
    for(l in c("P", "=", "X")) 
      keep[grep(l, cigarRead)]=FALSE
    mapq=mapq[keep]
    isize=isize[keep]
    posRead=as.integer(sampleData[[1]]$pos)[keep]
    cigarRead=sampleData[[1]]$cigar[keep]
    strandRead=sampleData[[1]]$strand[keep]
    seqRead=as.character(sampleData[[1]]$seq)[keep]
    qualRead=as.character(sampleData[[1]]$qual)[keep]
  qname=sampleData[[1]]$qname[keep]
    ###
    ### also, need to reconfigure cigar properly
    ###
    # do the most common one manually
    which=unique(
      c(grep("I",cigarRead),
        grep("D",cigarRead),
        grep("H",cigarRead),
        grep("S",cigarRead)
    ))
    # R doing weird things with removing negative vectors
    if (length(which) > 0) {
      x=cigarRead[-which]
    } else {
      x <- cigarRead
    }
    y=as.integer(substr(x,1,nchar(x)-1))
    z=tabulate(y)
    p=which.max(z)
    # now start filling in
    splitCigarRead=lapply(1:length(cigarRead),  function(x)  list(as.integer(p),"M"))
    # now - fix all the other ones!
    which=unique(c((1:length(cigarRead))[-which][y!=p],which))
    readLength=array(p,length(cigarRead))
    # figure out which ones are complicated
    splitCigarRead[which]=   lapply(cigarRead[which],function(c) {
      y=unlist(strsplit(c,""))
      t1=is.na(as.numeric(y))
      t2=(1:length(t1))[t1]
      t3=c(1,t2[-length(t2)]+1)
      n=length(t2)
      t5=array(0,n)
      t6=array("",n)
      for(i in 1:n)
      {
        t5[i]=substr(c,t3[i],t2[i]-1)
        t6[i]=substr(c,t2[i],t2[i])
      }
           return(list(as.integer(t5),t6))
    })
    ###
    ### deal with soft and hard clipped bases
    ###
    s <- sapply(splitCigarRead, function(x) sum(x[[2]]=="S"))
    h <- sapply(splitCigarRead, function(x) sum(x[[2]]=="H"))
    if(sum(s)>0) {
      if(useSoftClippedBases==TRUE) {
        # adjust position at the start appropriately if first is S
        w <- sapply(splitCigarRead, function(x) x[[2]][1]=="S")
        # get position of those that start with an S
        x <- sapply(splitCigarRead[w], function(x) x[[1]][1])
        # substract the starts back by that
        posRead[w] <- posRead[w] - x
        # change all the S's to M's
        splitCigarRead <- lapply(splitCigarRead, function(x) {
          x[[2]][x[[2]]=="S"]="M"
          return(x)
        })
      } else {
        # here we want to get rid of them
        # we need to effectively hard clip them
        # the sequence at the end is irrelevant
        w <- sapply(splitCigarRead, function(x) x[[2]][1]=="S")
        if (sum(w) > 0) {
          seqRead[w] <- sapply(which(w), function(i) {
            s1 <- seqRead[i]
            scr <- splitCigarRead[[i]]
            to_remove <- scr[[1]][1]
            return(substr(s1, 1+to_remove,nchar(s1)))
          })
          # now - remove them from the start
          splitCigarRead[w] <- lapply(splitCigarRead[w], function(x) {
            k <- x[[2]]!="S"
            return(list(x[[1]][k], x[[2]][k]))
          })          
        }
      }
    } # end of looking at soft clipped bases
    # hmm, looks like I can basically ignore hard clipped ones
    # the src function will skip them
    # and they aren't part of the alignment
    # get rest of lengths
    if (length(which) > 0) {
      readLength[which]=sapply(splitCigarRead[which],function(x) sum(x[[1]][FALSE==is.na(match(x[[2]],c("M","D")))]))
    }
    readStart=posRead # includes this first base
    readEnd=readStart+readLength-1 # includes final base
    ###
    ### push through c++ function
    ###
    lengthOfSplitCigarRead=as.integer(unlist(lapply(splitCigarRead,function(x) length(x[[1]])-1))) # 0 BASED
    ref=as.character(pos[,"REF"])
  alt=as.character(pos[,"ALT"]) 
    sampleReads=reformatReads(
      mapq = mapq,
      readStart = readStart,
      readEnd = readEnd,
      numberOfReads = as.integer(length(mapq)),
      T = T,
      L = L,
      seqRead = seqRead, 
      qualRead = qualRead,
      ref = ref,
      alt = alt,
      splitCigarRead = splitCigarRead,
      lengthOfSplitCigarRead = lengthOfSplitCigarRead,
      bqFilter = bqFilter,
      verbose = as.integer(0)
      )
    # check - if length 0
    if (length(sampleReads) > 0) {
      # output is list with 5 entries per read with SNPs
      # 1 = 0-based number of SNPs in read
      # 2 = 0
      # 3 = scaled bq, less than 0 if ref, greater than 0 if alt
      # 4 = 0-based indeces of reads
      # 5 = 0-based read it came from
      ###
      ### reassign to mate pairs
      ###
      wif=sapply(sampleReads,function(x) x[[5]])+1 # 1-based which read it came from
      qnameUnique=unique(qname)
      qnameInteger=match(qname,qnameUnique)
      a=qnameInteger[wif]
      b=unique(a)
      # want a vector - first one is the matches
      c1=match(b,a)
      z=a[c1]
      a[c1]=NA
      c2=match(z,a)
      # now - c1 is the position of the first read. c2 is the position of the second
      # for uniques - keep. for duplicates, merge!
      ###
      ### merge back together
      ###
      #
      sampleReads=lapply(1:length(c1),function(i)
      {
        if(is.na(c2[i])) {
          return(sampleReads[[c1[i]]][1:4])
        } else {
          a=sampleReads[[c1[i]]]
          b=sampleReads[[c2[i]]] # fixed nov 6 2015 - mice run 6.0 OK, 6.1 wrong
          return(list(as.integer(a[[1]]+b[[1]])+1,
                      0,
                      matrix(c(a[[3]],b[[3]]),ncol=1),
                      matrix(c(a[[4]],b[[4]]),ncol=1)))
        }
      }) # end of making sample reads
    }
    if(length(sampleReads)==0)
    {
      print(paste("WARNING - sample ",sampleNames[iBam]," has no informative reads. It is being given random reads. Consider removing from analysis",sep=""))
      sampleReads=lapply(1:5,function(i) list(0,matrix(as.integer(i),ncol=1,nrow=1),matrix(as.integer(i),ncol=1,nrow=1),matrix(as.integer(i),ncol=1,nrow=1)))
    } else {
      ###
      ### take a single, representative SNP from each - use this instead for position, etc.
      ###
      sampleReads=lapply(sampleReads,function(x) {
        if(length(x[[4]])>1) { x[[2]]=sample(x[[4]],1) } else { x[[2]]=x[[4]]} ; return(x)})
      t1=unlist(lapply(sampleReads,function(x) x[[2]]))
      sampleReads=sampleReads[order(t1)]
    }
  } # end of building given there exists information
  ###
  ### also - downsample here - can lead to problems when
  ### too high due to floating point errors
  ###
  sampleReads=downsample(sampleReads=sampleReads,iBam=iBam,downsampleToCov=downsampleToCov,sampleNames=sampleNames)  
  # could make rate map if so interested
  save(sampleReads,file=paste(inputdir,"sample.",iBam,".input.",regionName,".RData",sep=""))
}










shrinkReads=function(iSample,inputdir,originalRegionName, regionName,chr,tempdir,inRegionL,savedir)
{
  #
  # load file off disk
  #
  load(file=paste(inputdir,"sample.",iSample,".input.",originalRegionName,".RData",sep=""))
  #
  # figure out which intersect - keep
  #
  # snps intersected
  x=lapply(sampleReads,function(x) x[[4]]) # slow - next 3 fast
  snpPerRead=unlist(lapply(x,length))
  wrif=rep(1:length(sampleReads),snpPerRead)
  snpi=unlist(x)+1 # SNPs intersected - 1 based
  # keep those with SNPs completely within range
  w=  snpi>=inRegionL[1] & snpi<=inRegionL[2] # inRegionL is 1based
  readsToKeep=unique(wrif[w])
  sampleReads=sampleReads[readsToKeep]  
  #
  # now - remove those with SNPs outside
  #
  x=lapply(sampleReads,function(x) x[[4]])
  snpPerRead=unlist(lapply(x,length))
  wrif=rep(1:length(sampleReads),snpPerRead)
  snpi=unlist(x)+1 # SNPs intersected
  # keep those with SNPs completely within range
  w=  snpi<inRegionL[1] | snpi>inRegionL[2]
  readsToRemove=unique(wrif[w])
  if(length(readsToRemove)>0)
    sampleReads=sampleReads[-readsToRemove]
  #
  # now - change start and end
  #
  #a=0 # 0-based - what we keep
  #b=diff(inRegionL)-1
  sampleReads=lapply(sampleReads,function(x,inRegionL) {
    x[[2]]=x[[2]]-(inRegionL[1]-1)
    x[[4]]=x[[4]]-(inRegionL[1]-1)
    return(x)
  },inRegionL=inRegionL)
  save(sampleReads,file=paste(savedir,"sample.",iSample,".input.",regionName,".RData",sep=""))
  # OLD CODE - now I just remove anything which offends
  # buffer should be big enough to make this OK
  #  if(sum(x[[4]]<a) | sum(x[[4]]>b))
  #  {
  #    w=x[[4]]>=a & x[[4]]<=b
  #    x[[3]]=x[[3]][w]
  #    x[[4]]=x[[4]][w]      
  #    # choose new center - based on new smaller L
  #    LL=L[x[[4]]+1]
  #    # if length is 1 - choose it
  #    if(length(x[[4]])==1) x[[2]]=x[[4]]
  #    # if 2 - choose at random
  #    if(length(x[[4]])==2) x[[2]]=sample(x[[4]],1)
  #    # if>=3, choose based on minimizing distance
  #    if(length(x[[4]])>2)
  #      x[[2]]=x[[4]][which.min(sapply(1:length(LL),function(i) sum(abs(LL[i]-LL[-i]))))]
  #  }
  #  return(x)
  #
  #
  # save to temporary directory
  #
}




getCentral=function(LL)
{
  if(length(LL)==1) return(1)
  if(length(LL)==2) return(sample(2,1))
  if(length(LL)>=3) return(which.min(sapply(1:length(LL),function(i) sum(abs(LL[i]-LL[-i])))))
}




#subsetSNPsfile="/well/myers/rwdavies/converge/imputation.chr20.list.txt"
subsetSNPsFunction=function(N,subsetSNPsfile,regionName,tempdir,L,environment,nCores,outputdir)
{
  # load positions
  subsetList=scan(subsetSNPsfile)
  keep=is.na(match(L,subsetList))==FALSE
  # for those which are kept, do re-mapping
  keepL=array(NA,length(keep))
  keepL[keep]=(1:sum(keep))-1
  # loop through files
  f=function(iSample,tempdir,regionName,L,keep,outputdir)
  {
    load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
    # for loop through on sampleReads? fast enough?
    for(i in 1:length(sampleReads))
    {
      c=sampleReads[[i]][[3]]      
      d=sampleReads[[i]][[4]] # 0-based
      e=keep[d+1]
      if(sum(e==FALSE)==0) # keep everything!
      {
        # re-map only
        sampleReads[[i]][[2]]=keepL[sampleReads[[i]][[2]]+1]
        sampleReads[[i]][[4]]=keepL[sampleReads[[i]][[4]]+1]        
      } else {
        if(sum(e==TRUE)==0) # otherwise, remove!
        {
          sampleReads[[i]]=NA
        } else {
          # keepL is 0-based
          d2=keepL[matrix(d[e],ncol=1)+1] # 0-based
          c2=matrix(c[e],ncol=1)
          # get central position
          sampleReads[[i]]=list(sum(e)-1,d2[getCentral(L[d2+1])],c2,d2)
        }
      }
    }
    # remove those which are now NA
    sampleReads=sampleReads[sapply(sampleReads,length)==4]
    # done!
    save(sampleReads,file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
  }
  if(environment=="server")
  {
    out2=mclapply(1:N,FUN=f,tempdir=tempdir,regionName=regionName,L=L,keep=keep,mc.cores=nCores,outputdir=outputdir)
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, as.list(1:N), fun=f,tempdir=tempdir,regionName=regionName,L=L,keep=keep,outputdir=outputdir)
    save(out2,file=paste(outputdir,"debug/out2.RData",sep=""))
    stopCluster(cl)
  }
  #
  # done!
  #
  return(list(keep=keep))
}


  


### function to downsample BAMs as necessary
downsample=function(sampleReads,iBam,downsampleToCov,sampleNames)
{
  # first, figure out if reads need to be downsampled
  y=unlist(lapply(sampleReads,function(x) x[[2]]))
  # quick counts of each
  z=table(y)
  z1=as.numeric(names(z)[z>downsampleToCov])
  if(length(z1)>0)
  {
    # match each against
    z2=match(z1,y)
    z3=z2[is.na(z2)==FALSE]
    # get counts of each
    counts=as.numeric(z[match(z1,names(z))])
    # get positions of these elements
    # keep a list of those to remove
    toRemove=array(FALSE,length(sampleReads))
    for(i in 1:length(z1))
    {
      w=z3[i] # which read position to work on
      c=counts[i] # the counts
      range=w:(w+c-1)# the position range
      # downsample here
      toRemove[range]=TRUE
      toRemove[sample(range,downsampleToCov,replace=FALSE)]=FALSE
      # done!
    }
    # print out
    print(paste("WARNING - downsample sample ",sampleNames[iBam]," - ",sum(toRemove)," of ",length(sampleReads)," reads removed ",sep=""))
    # 
    sampleReads=sampleReads[toRemove==FALSE]
  }
  return(sampleReads)  
}



### split
splitReadsCompletely=function(N,nCores,tempdir,regionName,environment)
{
  # split into 48 cores
  f=function(iSample,tempdir,regionName) {
    load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
   #load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))    
    # split! 
    x3=unlist(lapply(sampleReads,function(x) x[[3]]))
    x4=unlist(lapply(sampleReads,function(x) x[[4]]))
    sr=lapply(1:length(x3),function(i) {
      list(0,x4[i],x3[i],x4[i])
    })
    sampleReads=sr[order(x4)]
    save(sampleReads,file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
  }
  if(environment=="server")
  {
    out2=mclapply(1:N,mc.cores=nCores,FUN=f,tempdir=tempdir,regionName=regionName)
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, 1:N, fun=f,tempdir=tempdir,regionName=regionName)
    stopCluster(cl)
  }
  # done
}



# convert scaled base quality to probabilities
convertScaledBQtoProbs=function(x)
{
  # for a matrix with 1 column, with base qualities, get the probabilities of the ref or alternate
  # recall < 0 is ref, >1 is alternate
  # output is a matrix, two columns
  n=nrow(x)
  o=array(0,c(n,2))
  e=10**(-abs(x)/10)
  # fill in
  w=cbind(1:n,2-as.integer(x<0))
  o[w]=1-e
  w[,2]=3-w[,2]
  o[w]=e*1/3
  return(o)
}


### build allele count matrix from input RData files
buildAlleleCount=function(T,N,nCores,regionName,tempdir,environment)
{
  # split into 48 cores
  x=seq(1,N,by=ceiling(N/nCores))
  x2=c(x[-1]-1,N)
  x3=as.list(1:nCores)
  for(i in 1:length(x2))
    x3[[i]]= c(x[i],x2[i])
  x3=x3[unlist(lapply(x3,length))>1] # make sure each one has at least two to load
  # do first loop - multicore this sucker
  f=function(i,T,tempdir,regionName) {
    alleleCount=array(0,c(T,3))
    for(j in i[1]:i[2])
    {
      load(file=paste(tempdir,"sample.",j,".input.",regionName,".RData",sep=""))
      # get positions and intensities
      a=unlist(sapply(sampleReads,function(x) x[[3]]))
      b=unlist(sapply(sampleReads,function(x) x[[4]]))
      bqProbs=convertScaledBQtoProbs(matrix(a,ncol=1))
      # y is numeric, z = integer counts (can have 0), 
      c1=increment2N(y=as.numeric(bqProbs[,1]),z=as.numeric(b),yT=as.integer(nrow(bqProbs)),xT=as.integer(T-1))
      c2=increment2N(y=as.numeric(bqProbs[,2]),z=as.numeric(b),yT=as.integer(nrow(bqProbs)),xT=as.integer(T-1))
      alleleCount[,1]=alleleCount[,1]+c2 # fixed nov 6 2015 - was backward
      alleleCount[,2]=alleleCount[,2]+c1+c2      
      # incrementN 
#      for(i in 1:length(sampleReads)) {
#        a=sampleReads[[i]]
#        # hmm - supposed we didn't need to convert? 
#        bqProbs=convertScaledBQtoProbs(a[[3]])
#        #bqProbs=cbind(a[[3]],a[[3]])
#        b=a[[4]]+1
#        # reference
#        alleleCount[b,1]= alleleCount[b,1]+bqProbs[,1]
#        # reference + alternate
#        alleleCount[b,2]= alleleCount[b,2]+bqProbs[,1]+bqProbs[,2]
#      }
    }
    return(list(alleleCount=alleleCount))
  }
  if(environment=="server")
  {
    out2=mclapply(x3,mc.cores=nCores,FUN=f,T=T,tempdir=tempdir,regionName=regionName) # end of out 2 - the multicore read in thingy
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, x3, fun=f,T=T,tempdir=tempdir,regionName=regionName)
    stopCluster(cl)
  }
  # now make single version
  alleleCount=array(0,c(T,3))
  #hapProbsSum=array(0,c(T,K))
  for(i in 1:length(x3))
  {
    alleleCount=alleleCount+out2[[i]][["alleleCount"]]
    #hapProbsSum=hapProbsSum+out2[[i]][["hapProbsSum"]]
  }
  alleleCount[,3]=alleleCount[,1]/alleleCount[,2]
  return(alleleCount)
}





### build allele count matrix from input RData files
downsampleToFraction=function(N,nCores,downsampleFraction,regionName,tempdir,environment)
{
  print(paste(c("Downsample to fraction - ",date()),collapse=""))
  if(downsampleFraction<0 | downsampleFraction>1) stop("ERROR - downsampleFraction must be between 0 and 1")
  # do first loop - multicore this sucker
  f=function(iSample,tempdir,regionName,downsampleFraction) {
    # load reads
    load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
    # subset
    keep=runif(length(sampleReads))<downsampleFraction
    # keep 1 no matter what
    if(sum(keep)==0) keep[sample(length(keep),1)]=TRUE
    sampleReads=sampleReads[keep]
    # save result back to disk
    save(sampleReads,file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
  }
  if(environment=="server")
  {
    out2=mclapply(1:N,mc.cores=nCores,FUN=f,tempdir=tempdir,regionName=regionName,downsampleFraction=downsampleFraction) # end of out 2 - the multicore read in thingy
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, 1:N, fun=f,tempdir=tempdir,regionName=regionName,downsampleFraction=downsampleFraction)
    stopCluster(cl)
  }
  return(NULL)
}



getOutputBlockRange <- function(
  sampleRange,
  outputBlockSize
) {
  d <- sampleRange[2]-sampleRange[1]
  if(outputBlockSize > d)
    outputBlockSize = d
  outputBlockRange <- c(
    sampleRange[1] - 1 + which((sampleRange[1]:sampleRange[2] -sampleRange[1])%%outputBlockSize==0),
      sampleRange[length(sampleRange)]
  )
  # subtract one from first entry 
  outputBlockRange[1] <- outputBlockRange[1] - 1
  if (outputBlockRange[length(outputBlockRange)] == outputBlockRange[length(outputBlockRange) - 1])
    outputBlockRange <- outputBlockRange[-length(outputBlockRange)]
  return(outputBlockRange)
}



getSampleRange <- function(
  N,
  nCores
) {
  x=seq(1,N,by=ceiling(N/nCores))
  x2=c(x[-1]-1,N)
  x3=as.list(1:nCores)
  for(i in 1:length(x2))
    x3[[i]]= c(x[i],x2[i])
  x3=x3[unlist(lapply(x3,length))>1] # make sure each one has at least two to load
  return(x3)
}


completeSampleIteration=function(N,tempdir,chr,K,T,priorCurrent,eHapsCurrent,alphaMatCurrent,sigmaCurrent,maxDifferenceBetweenReads,whatToReturn,Jmax,aSW=NA,bSW=NA,highCovInLow,iteration,method,expRate,minRate,maxRate,niterations,splitReadIterations,shuffleHaplotypeIterations,nCores,L,nGen,alphaMatThreshold,emissionThreshold,nHC,gen,outputdir,environment,pseudoHaploidModel,diploidModel,outputHaplotypeProbabilities,switchModelIteration,regionName,restartIterations,refillIterations,hapSumCurrent,outputBlockSize)
{
    
  #
  #
  #
  print(paste(c("Start of iteration ",iteration," - ",date()),collapse=""))
  if(is.na(match(iteration,restartIterations))==FALSE  )
    print(paste("Restart read probabilities ",date(),sep=""))
  #
  # first, we start by setting up some things depending on iteration
  # next, we run all samlpes with the current parameters
  # finally, we update all parameters
  #
  #
  #
  # initial set up based on iteration
  #
  Jmax=1000; # maximum number of SNPs per read (can be reduced early on)
  expRateL=expRate
  if(iteration<=10) # start slow - scale up later
    Jmax=as.integer(iteration-1) # slow-ish start, but scale up
  whatToReturn=as.integer(1)
  if(iteration==niterations) # final iteration - return everything - or splitting reads, need probabilities
    whatToReturn=as.integer(2)
  #
  #
  # set up switching iteration
  #
  # OLD ---- start
  #nbreaks=0
  #file=paste(tempdir,"switching.iteration",iteration-1,".RData",sep="")
  #if(file.exists(file))
  #  load(file)
  # OLD ---- end
  nbreaks=0
  if(is.na(match(iteration,shuffleHaplotypeIterations))==FALSE)
  {
    nSNP=100
    breaks=100*(0:(-1+floor(T/100))) + 51          
    nbreaks=length(breaks)-1
    fromMat=array(0,c(nbreaks,K,K))  
  }
  # set up splitting iteration
  nsplit=0
  if(is.na(match(iteration,splitReadIterations))==FALSE)
  {
    nsplit=1
  }
  #
  # 1 - run forward backwards
  # 2 - do updates
  # 3 - if necessary, fill into output files
  #
  x3 <- getSampleRange(N, nCores)
  #
  # normalize hapSumCurrent
  #
  hapSumCurrentL  =hapSumCurrent/rowSums(hapSumCurrent)
  #
  #
  # run multicore loop
  #
  f = function(sampleRange,tempdir,chr,K,T,priorCurrent,eHapsCurrent,alphaMatCurrent,sigmaCurrent,maxDifferenceBetweenReads,whatToReturn,Jmax,highCovInLow,iteration,method,nsplit,expRate,minRate,maxRate,nHC,gen,outputdir,pseudoHaploidModel,diploidModel,outputHaplotypeProbabilities,switchModelIteration,regionName,restartIterations,refillIterations,hapSumCurrentL,outputBlockSize)
  {
    #
    # determine whether its a special iteration
    #
    readsSplit=array(0,N)
    readsTotal=array(0,N)  
    nbreaks=0
    if(is.na(match(iteration,shuffleHaplotypeIterations))==FALSE)
    {
      nSNP=100
      breaks=100*(0:(-1+floor(T/100))) + 51      
      nbreaks=length(breaks)-1
      fromMat=array(0,c(nbreaks,K,K))  
    }
    # OLD --- start
    #file=paste(tempdir,"switching.iteration",iteration-1,".RData",sep="")
    #if(file.exists(file))
    #  load(file)
    # OLD --- end
    #
    # initialize refilling iteration
    #
    refillList=NA
    refillMatrixList=NA
    file=paste(outputdir,"RData/refillList.",regionName,".iteration.",iteration-1,".RData",sep="")    
    if(file.exists(file))
    {
      load(file)
      refillMatrixList=lapply(1:K,function(k) return(array(0,c(nrow(refillList[[k]]),K))))
      refillFromList=lapply(1:K,function(k) {
        # yes, we're using "1" for both
        # what we do is look at the start of the rare section
        # look what it was, then what it becomes
        from=sapply(refillList[[k]][,1]-50,function(x) max(1,x))
        to=sapply(refillList[[k]][,1]+50,function(x) min(T,x))
        return(list(from=from,to=to))
      })
    }
    #
    # initialize output matrices here
    #
    priorSum=array(0,K)
    alphaMatSum=array(0,c(T-1,K+1))
    gammaSum=array(0,c(T,K,2))
    hapSum=array(0,c(T,K))
    fromMat=array(0,c(3,3)) # dummy array
    if(nbreaks>0) fromMat=array(0,c(nbreaks,K,K))
    restartMatrixList=as.list(1:N)
    #
    # if final iteration - make count matrices
    #
    hweCount <- NA
    afCount <- NA
    infoCount <- NA
    if (iteration==niterations) {
      hweCount = array(0,c(T,3))
      infoCount = array(0,c(T,2))        
      afCount = array(0,T)
    }
    #
    # set output block size for VCF pasting
    #
    list_of_vcf_columns_to_out <- as.list(1:N)
    outputBlockRange <- getOutputBlockRange(
      sampleRange,
      outputBlockSize
    )
    #
    # 1 - run forward backwards
    # 2 - get update pieces
    # 3 - if necessary add to output
    #
    whatToReturnOriginal=whatToReturn    
    for(iSample in sampleRange[1]:sampleRange[2])
    {
      #
      #
      # 1 - run forward backwards
      #
      #
      whatToReturn=whatToReturnOriginal
      if(is.na(match(iSample,highCovInLow))==FALSE)
        whatToReturn=as.integer(2)
      # load both sample reads and read probabilities
      load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
      if(method=="pseudoHaploid")
      {
        if(iteration==1)
        {
          a1=0.4; b1=0.6;
          pRgivenH1=a1+(b1-a1)*runif(length(sampleReads))
          pRgivenH2=a1+(b1-a1)*runif(length(sampleReads))
          srp=unlist(lapply(sampleReads,function(x) x[[2]]))
          delta=0
          if(pseudoHaploidModel==6) {
            delta=array(0,length(srp)) # may or may not use? 
            delta=a1+(b1-a1)*runif(length(sampleReads))
          }
          save(srp,pRgivenH1,pRgivenH2,delta,file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))          
        }
        if(iteration>1)
          load(file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))
      }
      #
      # run single sample iteration here
      #
      # run either once (default) or more than once
      nor=1 # number of runs - 1 for normal diploid method
      # run both haplotypes
      if(method=="pseudoHaploid")
        nor=2
      fbsoL=as.list(1:nor)
      for(iNor in 1:nor)
      {
        if(method=="pseudoHaploid")
        {
          x=sigmaCurrent
          transMatRate=cbind(sigmaCurrent,1-sigmaCurrent)
          if(iNor==1)
            {    pRgivenH1L=pRgivenH1;    pRgivenH2L=pRgivenH2; deltaL=delta}
          if(iNor==2)
            {    pRgivenH1L=pRgivenH2;    pRgivenH2L=pRgivenH1; deltaL=1-delta}
          fbsoL[[iNor]]=forwardBackwardHaploid(
            sampleReads=sampleReads,
            nReads=as.integer(length(sampleReads)),
            Jmax=as.integer(Jmax),
            pi=priorCurrent,
            sigma=sigmaCurrent,
            pRgivenH1=pRgivenH1L,
            pRgivenH2=pRgivenH2L,
            pState=hapSumCurrentL,
            delta=delta,                                
            eHaps=eHapsCurrent,
            alphaMat=alphaMatCurrent,
            transMatRate=transMatRate,
            maxDifferenceBetweenReads=as.double(maxDifferenceBetweenReads),
            whatToReturn=whatToReturn,
            suppressOutput=as.integer(1),
            model=as.integer(pseudoHaploidModel))
        }
        if(method=="diploid")
        {
          x=sigmaCurrent
          transMatRate=cbind(x^2,x*(1-x),(1-x)^2)
          # note - number of SNPs determined by length of eHaps
          fbsoL[[iNor]]=forwardBackwardDiploid(
            sampleReads=sampleReads,
            nReads=as.integer(length(sampleReads)),
            pi=priorCurrent,
            eHaps=eHapsCurrent,
            alphaMat=alphaMatCurrent,
            sigma=sigmaCurrent,
            transMatRate=transMatRate,
            Jmax=Jmax,
            maxDifferenceBetweenReads=maxDifferenceBetweenReads,
            whatToReturn=whatToReturn,
            suppressOutput=as.integer(1),
            model=as.integer(diploidModel))
          # save everything to dev shm
        }
        #
        # get update pieces
        #
        which=fbsoL[[iNor]]$gammaUpdate[,1,1]>0
        jU=fbsoL[[iNor]][["jUpdate"]]
        gammaSum[which,,]=gammaSum[which,,]+fbsoL[[iNor]][["gammaUpdate"]][which,,]
        alphaMatSum=alphaMatSum+cbind(jU,1);
        priorSum=priorSum+fbsoL[[iNor]][["gammaUpdate"]][1,,2]
        if(method=="pseudoHaploid")
          fbsoL[[iNor]]$gammaK=fbsoL[[iNor]][["gamma"]]
      }
      #
      #
      #
      #
      # 2 - do various heuristics
      #
      #
      #
      #
      #
      # re-name gamma or get things needed for haplotype re-jigging
      #
      if(method=="diploid")
      {
        y=fbsoL[[iNor]][["gamma"]]
        gamma=y[,1:K]
        for(ii in 2:K)
          gamma=gamma+y[,1:K+(ii-1)*K]
        fbsoL[[iNor]]$gammaK=gamma
      }
      hapSum=hapSum + fbsoL[[iNor]]$gammaK
      if(nbreaks>0)
      {
        # for each sample, get the changes between every 100
        for(iNor in 1:nor)
        {
          hp=matrix(fbsoL[[1]]$gamma[breaks,],ncol=K)
          for(f in 1:nbreaks)
            fromMat[f,,]=fromMat[f,,] + hp[f,] %*% t(hp[f+1,])
        }
        ### OLD --- start
        #if(method=="pseudoHaploid") gamma=fbsoL[[1]]$gammaK+fbsoL[[2]]$gammaK
        #hp1=matrix(gamma[breaks1,],ncol=K)
        #hp2=matrix(gamma[breaks2,],ncol=K)
        #for(f in 1:nbreaks)
        #  fromMat[f,,]=fromMat[f,,] + hp1[f,] %*% t(hp2[f,])
        ### OLD --- end
      }
      #
      # now - consider whether to re-start - ie fix gammas 
      #
      if(is.na(match(iteration,restartIterations))==FALSE)
      {
        out=restartSample(iSample=iSample,regionName=regionName,tempdir=tempdir,fbsoL=fbsoL,T=T,eHapsCurrent=eHapsCurrent,sigmaCurrent=sigmaCurrent,alphaMatCurrent=alphaMatCurrent,Jmax=Jmax,priorCurrent=priorCurrent,maxDifferenceBetweenReads=maxDifferenceBetweenReads,outputdir=outputdir,diploidModel=diploidModel)
        restartMatrixList[[iSample]]=out$m
        fbsoL=out$fbsoL
        # note that after this, we need a re-start of the read probabilities, but this follows immediately
        # note that the output tells you what happened
      }
      #
      #
      #
      # update pseudo diploid probabilities
      #
      if(method=="pseudoHaploid")      
      {
        r=pseudoHaploidUpdate(pRgivenH1=pRgivenH1,pRgivenH2=pRgivenH2,fbsoL=fbsoL,pseudoHaploidModel=pseudoHaploidModel,srp=srp,delta=delta)
        pRgivenH1=r$pRgivenH1
        pRgivenH2=r$pRgivenH2
        delta=r$delta
        save(srp,pRgivenH1,pRgivenH2,delta,file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))
        #m=pseudoHaploidFindRegion(iSample=iSample,regionName=regionName,tempdir=tempdir)
      }
      #
      # check high coverage samples if applicable
      #
      if(is.na(match(iSample,highCovInLow))==FALSE)
      {
        dosage=array(0,T)
        for(iNor in 1:nor)
          dosage=dosage+(rowSums(fbsoL[[iNor]]$gammaK * eHapsCurrent))
        if(nor==2) dosage=dosage/2
        save(dosage,file=paste(tempdir,"sample.",iSample,".input.",regionName,".dosage.RData",sep=""))
        if(match(iSample,highCovInLow)==1 & 1==0)
        {
          g1=fbsoL[[1]]$eMatHap
          g2=fbsoL[[2]]$eMatHap
          g1A=apply(g1,1,which.max)
          g2A=apply(g2,1,which.max)
          print(table(g1A,g2A))
          # table
          print(table(round(pRgivenH1,1),round(pRgivenH2,1)))
        }
      } # end of check on high coverage sample
      #
      # do read splitting if the correct iteration
      #
      if(nsplit>0)
      {
        if(method=="pseudoHaploid") gamma=fbsoL[[1]]$gammaK+fbsoL[[2]]$gammaK
        a=findRecombinedReadsPerSample(fbsoL=fbsoL,gamma=gamma,eHapsCurrent=eHapsCurrent,K=K,L=L,iSample=iSample,verbose=FALSE,sampleReads=sampleReads,tempdir=tempdir,N=N,regionName=regionName)
        readsTotal[iSample]=a$readsTotal
        readsSplit[iSample]=a$readsSplit
      } # end of check on whether
      #
      #
      #
      #
      #
      # 3 - output as necessary
      #
      #
      #
      if(iteration==niterations)
      {
        # get allele counts for HWE, info calcs, AF
        # get most likely genotype - add to count
        for(i in 1:length(fbsoL)) {
          if ( method == "diploid") {
            gp <- fbsoL[[1]]$genProbs
          } else {
            gp <- array(0, c(T, 3))
            g10 <- rowSums(fbsoL[[1]]$gammaK * (1-eHapsCurrent))
            g20 <- rowSums(fbsoL[[2]]$gammaK * (1-eHapsCurrent))
            gp[,1] <- g10 * g20
            gp[,2] <- g10 * (1-g20) + (1-g10) * g20
            gp[,3] <- (1-g10) * (1-g20)
          }
        }
        # info counts
        eij <- gp[,2] + 2 * gp[,3]
        fij <- gp[,2] + 4 * gp[,3]  
        infoCount[,1] <- infoCount[,1] + eij
        infoCount[,2] <- infoCount[,2] + (fij - eij**2)
        # do counts for HWE
        w <- get_max_gen_rapid(gp)
        hweCount[w] <- hweCount[w]+1
        # get counts for allele frequency
        afCount <- afCount + (gp[,2] + 2*gp[,3]) / 2
        #
        #
        # add column to VCF, and possibly write block to disk
        list_of_vcf_columns_to_out[[iSample]] <- make_column_of_vcf(gp)
        i_core <- match(sampleRange[1], sapply(x3, function(x) x[[1]]))        
        iBlock <- match(iSample, outputBlockRange)
        if (iBlock > 1 & is.na(iBlock)==FALSE) {
          iSample_list <- (outputBlockRange[iBlock-1]+1):outputBlockRange[iBlock]            
          write_block_of_vcf(i_core, iBlock, list_of_vcf_columns_to_out, iSample_list = iSample_list, outputdir, regionName)
          list_of_vcf_columns_to_out <- as.list(1:N)
        }
      } # end of check on whether its the final iteration and to output
    } # end of sample loop
    #
    # end loop on samples being processed
    #
    return(
      list(
        alphaMatSum = alphaMatSum,
        gammaSum = gammaSum,
        priorSum = priorSum,
        hapSum = hapSum,
        fromMat = fromMat,
        readsSplit = readsSplit,
        readsTotal = readsTotal,
        restartMatrixList = restartMatrixList,
        refillMatrixList = refillMatrixList,
        infoCount = infoCount,
        afCount = afCount,
        hweCount = hweCount
      )
    )
  } # end of out 2 - the multicore read in thingy
  #
  # now - run the above in either a server or cluster environment
  #
  if(environment=="server")
  {
    out2=mclapply(x3,mc.cores=nCores,tempdir=tempdir,chr=chr,K=K,T=T,priorCurrent=priorCurrent,eHapsCurrent=eHapsCurrent,alphaMatCurrent=alphaMatCurrent,sigmaCurrent=sigmaCurrent,maxDifferenceBetweenReads=maxDifferenceBetweenReads,whatToReturn=whatToReturn,Jmax=Jmax,highCovInLow=highCovInLow,iteration=iteration,method=method,nsplit=nsplit,expRate=expRate,minRate=minRate,maxRate=maxRate,gen=gen,outputdir=outputdir,pseudoHaploidModel=pseudoHaploidModel,diploidModel=diploidModel,outputHaplotypeProbabilities=outputHaplotypeProbabilities,switchModelIteration=switchModelIteration,regionName=regionName,restartIterations=restartIterations,refillIterations=refillIterations,hapSumCurrentL=hapSumCurrentL,outputBlockSize=outputBlockSize,FUN=f)
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, x3, fun=f,tempdir=tempdir,chr=chr,K=K,T=T,priorCurrent=priorCurrent,eHapsCurrent=eHapsCurrent,alphaMatCurrent=alphaMatCurrent,sigmaCurrent=sigmaCurrent,maxDifferenceBetweenReads=maxDifferenceBetweenReads,whatToReturn=whatToReturn,Jmax=Jmax,highCovInLow=highCovInLow,iteration=iteration,method=method,nsplit=nsplit,expRate=expRate,minRate=minRate,maxRate=maxRate,gen=gen,outputdir=outputdir,pseudoHaploidModel=pseudoHaploidModel,diploidModel=diploidModel,outputHaplotypeProbabilities=outputHaplotypeProbabilities,switchModelIteration=switchModelIteration,regionName=regionName,restartIterations=restartIterations,refillIterations=refillIterations,hapSumCurrentL=hapSumCurrentL,outputBlockSize=outputBlockSize)
    stopCluster(cl)
  }



  #
  #
  #
  #
  # done running samples
  # first, check if there was an error, print it
  #
  #
  #
  #
  x=unlist(lapply(out2,length))
  if(sum(x!=12)>0)
    print(out2[[which.min(x)]])
  # print out first if its an error
  #
  # now do the second part of the updating here - collapse results
  #
  # now make single version
  priorSum=array(0,K)
  alphaMatSum=array(0,c(T-1,K+1))
  gammaSum=array(0,c(T,K,2))
  hapSum=array(0,c(T,K))
  restartMatrixList=as.list(1:N)
  # note - already defined fromMat before the function f above
  if(nsplit>0) {
    readsTotal=array(0,N)
    readsSplit=array(0,N)    
  }
  # sum and sum
  for(i in 1:length(x3))
  {
    priorSum=priorSum+out2[[i]][["priorSum"]]
    alphaMatSum=alphaMatSum+out2[[i]][["alphaMatSum"]]
    gammaSum=gammaSum+out2[[i]][["gammaSum"]]
    hapSum=hapSum + out2[[i]][["hapSum"]]
    if(nbreaks>0) fromMat=fromMat+out2[[i]][["fromMat"]]
    if(nsplit>0) {
      readsTotal=readsTotal+out2[[i]][["readsTotal"]]
      readsSplit=readsSplit+out2[[i]][["readsSplit"]]
    }
    restartMatrixList[x3[[i]][1]:x3[[i]][2]]=out2[[i]]$restartMatrixList[x3[[i]][1]:x3[[i]][2]]
  }
  # now split apart alphaMatSum
  nAlphaMatSum=alphaMatSum[,K+1]
  alphaMatSum=alphaMatSum[,-(K+1)]
  # normalize prior
  priorSum=priorSum/sum(priorSum)
  # note - it is possible in simulations not to have sampled the first SNP
  # in which case there is no posterior information, so set to random
  if(is.na(priorSum[1]))
    priorSum=rep(1/K,K)
  priorSum[priorSum<0.0001]=0.0001
  priorSum=priorSum/sum(priorSum)
  # bound above and below by a factor of 1000 
  # normalize sigma
  sigmaSum=rowSums(alphaMatSum)/N/2
  #sigmaSum=rowSums(alphaMatSum)/2  /nAlphaSum
  sigmaSum=exp(-sigmaSum) # 100 is useless after start if its cant update
  # NA values - set to expectation
  which=is.na(sigmaSum)
  dl=diff(L)
  #if(is.na(aSW)==FALSE) dl=diff(L[aSW:bSW+1])
  sigmaSum[which]=exp(-nGen*expRate/100/1000000*dl)[which]
  # morgans per SNP assuming T=100, 0.5 cM/Mb
  #div=max(10**(10-iteration),1) # more harsh at start
  div=1
  x1=exp(-nGen * minRate * dl/100/1000000) # lower
  x2=exp(-nGen * maxRate * dl/100/1000000) # upper
  # recombination average rate
  #print(paste("Mean recombination - ",mean(-(log(sigmaSum))*100*1e6)/diff(range(L)),sep=""))
  #print(paste("Recombination - number below minimum ",sum(sigmaSum>x1),sep=""))
  #print(paste("Recombination - number above maximum ",sum(sigmaSum<x2),sep=""))  
  # we estiamte the compound parameter T * sigma_t
  # so we need to use nGen to bound apppropriately
  sigmaSum[sigmaSum>x1]=x1[sigmaSum>x1]
  sigmaSum[sigmaSum<x2]=x2[sigmaSum<x2]
  # make sure each alphaMatSum has some minimum value - say 1
  # prevent the use of this being overriden by, say, one sample trying to switch
  # normalize alphaMatSum
  alphaMatSum=alphaMatSum+1 # add a bit to smooth things out
  alphaMatSum=alphaMatSum/rowSums(alphaMatSum)
  # 
  gammaSum=gammaSum[,,1]/gammaSum[,,2]
  # fix with min, max value
  gammaSum[rowSums(is.na(gammaSum))>0,]=0.5 # nov 6 2015 - not sure this matters? 
  gammaSum[gammaSum>(1-emissionThreshold)]=(1-emissionThreshold)
  gammaSum[gammaSum<emissionThreshold]=emissionThreshold
  # fix alphaMatSum as well
  alphaMatSum[alphaMatSum<alphaMatThreshold]=0
  alphaMatSum=    (1-(alphaMatThreshold*rowSums(alphaMatSum==0))) * alphaMatSum/ rowSums(alphaMatSum)
  alphaMatSum[alphaMatSum==0]=alphaMatThreshold
  # fix if NA
  t1=rowSums(is.na(alphaMatSum))>0
  if(sum(t1)>0) print("WARNING - alphaMat update contains NA")
  alphaMatSum[rowSums(is.na(alphaMatSum))>0,]=0.25
  #
  #
  #
  hwe <- NA
  info <- NA
  estimatedAlleleFrequency <- NA
  if(iteration == niterations) {
    infoCount <- array(0,c(T,2))   
    hweCount <- array(0,c(T,3))
    afCount <- array(0,T)
    for(i in 1:length(out2)) {
      infoCount <- infoCount + out2[[i]]$infoCount
      hweCount <- hweCount + out2[[i]]$hweCount
      afCount <- afCount + out2[[i]]$afCount
    }
    thetaHat <- infoCount[,1] / 2 / N
    denom <- 2 * N * thetaHat * (1-thetaHat)
    info <- 1 - infoCount[,2] / denom
    estimatedAlleleFrequency <- afCount / N
    hwe <- generate_hwe_on_counts(hweCount, T, nCores)
  }
  #
  #
  # shuffle haplotype iterations
  #
  #
  #
  # look at switching
  #
  if(nbreaks>0)
  {
    out=getBetterSwitchesSimple(fromMat=fromMat,nbreaks=nbreaks,K=K,T=T,eHapsFuture=gammaSum,alphaMatFuture=alphaMatSum,iteration=iteration)
    gammaSum=out$eHapsFuture
    alphaMatSum=out$alphaMatFuture
  }
  ### OLD ------- start
  if(1==0) {
  if(is.na(match(iteration,shuffleHaplotypeIterations))==FALSE)
    x=getBetterSwitchesRound1(iteration=iteration,tempdir=tempdir,nGen=nGen,expRate=expRate,sigmaFuture=sigmaSum,T=T,maxRate=maxRate,L=L)
  # if output file exists and correct iteration, run round 2
  if(nbreaks>0)
  {
    out=getBetterSwitchesRound2(fromMat=fromMat,K=K,nGen=nGen,expRate=expRate,L=L,eHapsFuture=gammaSum,sigmaFuture=sigmaSum,alphaMatFuture=alphaMatSum,iteration=iteration,tempdir=tempdir)
    gammaSum=out$eHapsFuture
    sigmaSum=out$sigmaFuture
    alphaMatFuture=out$alphaMatFuture
  }
  }
  ### OLD ------- end
  #
  # look at refilling - round 1
  #
  if(is.na(match(iteration,refillIterations))==FALSE)
  {
    print(paste("Iteration - ",iteration," - refill infrequently used haplotypes",sep=""))
    out=refillSimple(hapSum=hapSum,T=T,K=K,eHapsCurrent=gammaSum,N=N)
    eHapsCurrent=out$eHapsCurrent
    noise=0.2 # add in noise so things aren't exact
    eHapsCurrent=noise * array(runif(T*K),c(T,K)) + (1-noise)*eHapsCurrent
    priorSum=noise * rep(1/K,K) + (1-noise)*priorSum # restart these as well
    gammaSum=eHapsCurrent
    # add in noise as well
  }
  #
  #
  #
  # POSSIBLY DELETE BELOW
  #
  #
  #
  if(1==0) # is.na(match(iteration-1,refillIterations))==FALSE)
  {
    load(file  =paste(outputdir,"RData/refillList.",regionName,".iteration.",iteration-1,".RData",sep=""))
    # have refillList - where to refill
    refillMatrixList=lapply(1:K,function(k)  array(0,c(nrow(refillList[[k]]),K)))
    for(i in 1:length(out2))
      for(k in 1:K)
        refillMatrixList[[k]]=refillMatrixList[[k]]+out2[[i]]$refillMatrixList[[k]]
    refillFromList=lapply(1:K,function(k) {
      from=sapply(refillList[[k]][,1]-50,function(x) max(1,x))
      to=sapply(refillList[[k]][,1]+50,function(x) min(T,x))
      return(list(from=from,to=to))
    })
    refillList2=lapply(refillList,function(x) return(matrix(cbind(x,-1,-1))))
    #
    # now - start refilling
    # 
    for(k in 1:K)
    {
      # don't do anything if it stays
      m2=refillList[[k]]      
      x=apply(refillMatrixList[[k]],1,which.max)
      toSwitch=x!=k
      if(m2[1,1]==1) { toSwitch[1]=TRUE; x=which.max(hapSum[1,]) }
      m2=cbind(k,hapSum[cbind(refillFromList[[k]][[1]],k)],toSwitch,hapSum[cbind(refillFromList[[k]][[1]],x)])
      if(sum(toSwitch>0))
      {
        # these are the ones to refill, using x
        m=matrix(refillList[[k]][toSwitch,],ncol=3)
        for(j in 1:nrow(m))
          eHapsCurrent[m[j,1]:m[j,2],k]=eHapsCurrent[m[j,1]:m[j,2],x[j]]
      } # end of whether there are any to swich
    } # end of loop on which to refill
    # now - also - add noise back into the system
    # otherwise, no updates
    noise=0.2
    eHapsCurrent=noise * array(runif(T*K),c(T,K)) + (1-noise)*eHapsCurrent
    priorCurrent=noise * rep(1/K,K) + (1-noise)*priorCurrent # restart these as well
  } # end of refill iteration 
  #
  # look at splitting
  #
  if(nsplit>0)
  {
    print(paste("split reads - average ",round(mean(readsSplit))," % ",round(100*mean(readsSplit)/mean(readsTotal),3),",",date(),sep=""))
    save(readsTotal,readsSplit,file  =paste(outputdir,"RData/splitReads.",regionName,".iteration.",iteration,".RData",sep=""))
  }
  #
  # finally, if we have high coverage samples, run them
  #
  if(length(highCovInLow)>0)
  {
    currentR2=getR2DuringEMAlgorithm(highCovInLow,gen=gen,K=K,T=T,iteration=iteration,runVersion=runVersion,chr=chr,tempdir=tempdir,nHC=nHC,regionName=regionName)
    # output!
    write.table(matrix(c(iteration,round(currentR2,4),date()),nrow=1),file=paste(outputdir,"RData/interim.r2",chr,".txt",sep=""),row.names=FALSE,col.names=FALSE,sep="\t",quote=FALSE,append=iteration>1)
    print(paste(c(iteration,round(currentR2,3),date()),collapse=" - "))
  }
  #
  # also, if its a restart iteration, print out some stats
  #
  if(is.na(match(iteration,restartIterations))==FALSE)
  {
    restartMatrix=cbind(rep(1:N,sapply(restartMatrixList,nrow)),matrix(unlist(lapply(restartMatrixList,t)),ncol=5,byrow=TRUE))
    save(restartMatrix,file=paste(outputdir,"RData/restartMatrix.iteration.",iteration,".",regionName,".RData",sep=""))
  }
  # pump out some stats - av number per sample, etc
  #
  # done! 
  #
  return(
    list(
      eHapsFuture = gammaSum,
      sigmaFuture = sigmaSum,
      priorFuture = priorSum,
      alphaMatFuture = alphaMatSum,
      hapSum = hapSum,
      info = info,
      estimatedAlleleFrequency = estimatedAlleleFrequency,
      hwe = hwe
    )
  )
}






# for a sample, first, find regions where the two haplotypes are nearly the same
# next, loop over these regions. run diploid forward backward algorithm
# once that's done, re-do probabilities of the two haplotypes
restartSample=function(iSample,regionName,tempdir,fbsoL,T,eHapsCurrent,sigmaCurrent,alphaMatCurrent,Jmax,priorCurrent,maxDifferenceBetweenReads,outputdir,diploidModel)
{
  #
  # first, find the regions we want to run this on
  #
  m=pseudoHaploidFindRegion(iSample=iSample,regionName=regionName,tempdir=tempdir)
  m=cbind(matrix(m,ncol=3),NA,NA)
  if(sum(m[,3]==1)==0) return(list(m=m,fbsoL=fbsoL))
  #
  # otherwise - continue
  #
  xx=(1:nrow(m))[m[,3]==1]
  for(iR in 1:length(xx))
  {
    start=m[xx[iR],1]
    end=m[xx[iR],2]
    #
    # now - over this region - re-run diploid
    # get a viterbi path back
    #
    # determine a range of SNPs to run over
    #
    load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))  
    load(file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))  
    #
    # buffer another 50 reads to either side
    #
    start=max(1,start-50)
    end=min(length(sampleReads),end+50)
    TR=range(unlist(lapply(sampleReads[start:end],function(x) x[[4]])))+1 # 1-based SNPs to use
    #
    # subset reads
    #
    sampleReads=sampleReads[start:end]
    sampleReads=lapply(sampleReads,function(x) {
      # remember - these are 0-based
      x[[2]]=x[[2]]-(TR[1]-1)
      x[[4]]=x[[4]]-(TR[1]-1)      
      return(x)
    })
    #
    # forward backward diploid here
    #
    x=sigmaCurrent[TR[1]:(TR[2]-1)]
    transMatRate=cbind(x^2,x*(1-x),(1-x)^2)
    fbd=forwardBackwardDiploid(
      sampleReads=sampleReads,
      nReads=as.integer(length(sampleReads)),
      pi=priorCurrent,
      eHaps=eHapsCurrent[TR[1]:TR[2],],
      alphaMat=alphaMatCurrent[TR[1]:(TR[2]-1),],
      sigma=x,
      transMatRate=transMatRate,
      Jmax=Jmax,
      maxDifferenceBetweenReads=maxDifferenceBetweenReads,
      whatToReturn=as.integer(1),
      suppressOutput=as.integer(1),
      model=as.integer(diploidModel))
    #
    # compare to haploid run, try to properly integrate
    #
    a1=fbsoL[[1]]$gamma[c(TR[1],TR[2]),]
    a2=fbsoL[[2]]$gamma[c(TR[1],TR[2]),]
    b1=fbd$gamma[c(1,TR[2]-TR[1]+1),]
    #
    # simplest - argmax. symmetry shouldn't be a problem as argmax has non random tie breaks
    #
    z=apply(fbd$gamma,1,which.max)
    k1=z%%K
    k1[k1==0]=K
    k2=(z-k1)/K+1
    #
    # try to determine right way to switch together
    #
    startD=c(k1[1],k2[1])
    endD=c(k1[length(k1)],k2[length(k2)])
    #
    # now - either determine whether c(hap1 ,h
    #
    # first option being bigger - means k1 matches hap1, k2 matches hap2
    # otherwhise, opposite
    wS=c(a1[1,startD[1]]+a2[1,startD[2]],a2[1,startD[1]]+a1[1,startD[2]])
    wSX=which.max(wS)
    # now look at the end
    wE=c(a1[2,endD[1]]+a2[2,endD[2]],a2[2,endD[1]]+a1[2,endD[2]])
    wEX=which.max(wE)    
    #
    # if max or wE and wS are the same, we're OK, no switch required
    #
    # no matter what, push updates to gamma, based on new states
    # wS is all that matters
    #
    for(iHap in 1:2)
    {
      if(wSX==1 & iHap==1) kL=k1
      if(wSX==1 & iHap==2) kL=k2
      if(wSX==2 & iHap==1) kL=k2
      if(wSX==2 & iHap==2) kL=k1
      p=0.1 # "error" split over other states
      q=array(runif((1+diff(TR))*K),c(1+diff(TR),K)) # the new gammas for that region
      q=p * q/rowSums(q)
      q[cbind(1:nrow(q),kL)]=(1-p) + q[cbind(1:nrow(q),kL)]
      # changes to gamma for 1 are based on k1
      fbsoL[[iHap]]$gamma[TR[1]:TR[2],]=q
    }
    if(wSX!=wEX & TR[2]!=T) # if they are different, and there's something left to do
    {
      a1=fbsoL[[1]]$gamma[(TR[2]+1):T,]
      a2=fbsoL[[2]]$gamma[(TR[2]+1):T,]      
      fbsoL[[2]]$gamma[(TR[2]+1):T,]=a2
      fbsoL[[1]]$gamma[(TR[2]+1):T,]=a1
    }
    #
    # out of interest, record whether a flip has been done
    #
    m[xx[iR],4:5]=c(wSX,wEX)
    #
  } # end of loop over region
  #
  # done!
  #
  return(list(m=m,fbsoL=fbsoL))
}






# for a sample, first, find regions where the two haplotypes are nearly the same
# next, loop over these regions. run diploid forward backward algorithm
# once that's done, re-do probabilities of the two haplotypes
restartSampleDEBUG=function(iSample,regionName,tempdir,fbsoL,T,eHapsCurrent,sigmaCurrent,alphaMatCurrent,Jmax,priorCurrent,maxDifferenceBetweenReads,outputdir)
{
  #
  # first, find the regions we want to run this on
  #
  debugFile=paste(outputdir,"debug/sample.",iSample,".",regionName,".txt",sep="") #XXX
  cat("In restartSample sample",iSample,"\n",file=debugFile) #XXX
  m=pseudoHaploidFindRegion(iSample=iSample,regionName=regionName,tempdir=tempdir)
  m=cbind(matrix(m,ncol=3),NA,NA)
  cat("Generated region sample",iSample,"\n",file=debugFile,append=TRUE) #XXX
  save(m,file=paste(outputdir,"debug/sample.m.",iSample,".RData",sep="") ) #XXX
  load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))  
  load(file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))
  save(pRgivenH1,pRgivenH2,srp,sampleReads,file=paste(outputdir,"debug/sample.input.",iSample,".",regionName,".RData",sep="") )
  if(sum(m[,3]==1)==0) return(list(m=m,fbsoL=fbsoL))
  #
  # otherwise - continue
  #
  xx=(1:nrow(m))[m[,3]==1]
  for(iR in 1:length(xx))
  {
    #
    #
    #
    cat("Generated region sample",iSample," iR ",iR,"\n",file=debugFile,append=TRUE) #XXX
    #
    # 
    #
    start=m[xx[iR],1]
    end=m[xx[iR],2]
    #
    # now - over this region - re-run diploid
    # get a viterbi path back
    #
    # determine a range of SNPs to run over
    #
    cat("Load\n",file=debugFile,append=TRUE) #XXX    
    load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))  
    load(file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))  
    #
    # buffer another 50 reads to either side
    #
    cat("Buffer\n",file=debugFile,append=TRUE) #XXX        
    start=max(1,start-50)
    end=min(length(sampleReads),end+50)
    TR=range(unlist(lapply(sampleReads[start:end],function(x) x[[4]])))+1 # 1-based SNPs to use
    #
    # subset reads
    #
    cat("Subset\n",file=debugFile,append=TRUE) #XXX            
    sampleReads=sampleReads[start:end]
    sampleReads=lapply(sampleReads,function(x) {
      # remember - these are 0-based
      x[[2]]=x[[2]]-(TR[1]-1)
      x[[4]]=x[[4]]-(TR[1]-1)      
      return(x)
    })
    #
    # forward backward diploid here
    #
    cat("Forward backwards\n",file=debugFile,append=TRUE) #XXX   
    x=sigmaCurrent[TR[1]:(TR[2]-1)]
    transMatRate=cbind(x^2,x*(1-x),(1-x)^2)
    fbd=forwardBackwardDiploid(
      sampleReads=sampleReads,
      nReads=as.integer(length(sampleReads)),
      pi=priorCurrent,
      eHaps=eHapsCurrent[TR[1]:TR[2],],
      alphaMat=alphaMatCurrent[TR[1]:(TR[2]-1),],
      sigma=x,
      transMatRate=transMatRate,
      Jmax=Jmax,
      maxDifferenceBetweenReads=maxDifferenceBetweenReads,
      whatToReturn=as.integer(1),
      suppressOutput=as.integer(1))
    #
    # compare to haploid run, try to properly integrate
    #
    a1=fbsoL[[1]]$gamma[c(TR[1],TR[2]),]
    a2=fbsoL[[2]]$gamma[c(TR[1],TR[2]),]
    b1=fbd$gamma[c(1,TR[2]-TR[1]+1),]
    #
    # simplest - argmax. symmetry shouldn't be a problem as argmax has non random tie breaks
    #
    cat("Argmax\n",file=debugFile,append=TRUE) #XXX                    
    z=apply(fbd$gamma,1,which.max)
    k1=z%%K
    k1[k1==0]=K
    k2=(z-k1)/K+1
    #
    # try to determine right way to switch together
    #
    startD=c(k1[1],k2[1])
    endD=c(k1[length(k1)],k2[length(k2)])
    #
    # now - either determine whether c(hap1 ,h
    #
    # first option being bigger - means k1 matches hap1, k2 matches hap2
    # otherwhise, opposite
    cat("Start and end\n",file=debugFile,append=TRUE) #XXX                        
    wS=c(a1[1,startD[1]]+a2[1,startD[2]],a2[1,startD[1]]+a1[1,startD[2]])
    wSX=which.max(wS)
    # now look at the end
    wE=c(a1[2,endD[1]]+a2[2,endD[2]],a2[2,endD[1]]+a1[2,endD[2]])
    wEX=which.max(wE)    
    #
    # if max or wE and wS are the same, we're OK, no switch required
    #
    # no matter what, push updates to gamma, based on new states
    # wS is all that matters
    #
    cat("Re-do1\n",file=debugFile,append=TRUE) #XXX                            
    for(iHap in 1:2)
    {
      if(wSX==1 & iHap==1) kL=k1
      if(wSX==1 & iHap==2) kL=k2
      if(wSX==2 & iHap==1) kL=k2
      if(wSX==2 & iHap==2) kL=k1
      p=0.1 # "error" split over other states
      q=array(runif((1+diff(TR))*K),c(1+diff(TR),K)) # the new gammas for that region
      q=p * q/rowSums(q)
      q[cbind(1:nrow(q),kL)]=(1-p) + q[cbind(1:nrow(q),kL)]
      # changes to gamma for 1 are based on k1
      fbsoL[[iHap]]$gamma[TR[1]:TR[2],]=q
    }
    cat("Re-do2\n",file=debugFile,append=TRUE) #XXX                                
    if(wSX!=wEX & TR[2]!=T) # if they are different, and there's something left to do
    {
      a1=fbsoL[[1]]$gamma[(TR[2]+1):T,]
      a2=fbsoL[[2]]$gamma[(TR[2]+1):T,]      
      fbsoL[[2]]$gamma[(TR[2]+1):T,]=a2
      fbsoL[[1]]$gamma[(TR[2]+1):T,]=a1
    }
    #
    # out of interest, record whether a flip has been done
    #
    cat("Record\n",file=debugFile,append=TRUE) #XXX                                    
    m[xx[iR],4:5]=c(wSX,wEX)
    #
  } # end of loop over region
  cat("Done sample",iSample,"\n",file=debugFile,append=TRUE) #XXX
  #
  # done!
  #
  return(list(m=m,fbsoL=fbsoL))
}



  

pseudoHaploidFindRegion=function(iSample,regionName,tempdir)
{
  # in this function, we define underperforming regions
  #srp,pRgivenH1,pRgivenH2
    load(file=paste(tempdir,"sample.",iSample,".readProbs.",regionName,".RData",sep=""))
    # define an underperforming region as a set of >100 SNPs with av <0.0001 difference
    n=100 # region to check over
    n2=100 # when a region is smaller than this, remove it
    x=abs(pRgivenH1-pRgivenH2)
    y=cumsum(x)
    p=y[-c(1:(n-1))]-y[-(length(y)-(n-2):0)]
    w=p<1
    # turn into a set of regions
    z=array(FALSE,length(pRgivenH1))
    z[c(array(FALSE,n-1),w)]=TRUE
    z[c(w,array(FALSE,n-1))]=TRUE
    #z=array(FALSE,length(pRgivenH1))
    #z[1000:1200]=TRUE
    #z[1250:1600]=TRUE
    #z[1670:1800]=TRUE            
    #z[3000:3500]=TRUE    
    start=c(1,(1:length(z))[diff(z)!=0]+1)
    end=c((1:length(z))[diff(z)!=0],length(pRgivenH1))
    state=z[start]
    # recap
    # have length(pRgivenH1) reads
    # those are now in regions, which start at "start" and end at "end"
    # inside those regions, they all have the same state as "state"
    #
    # if there is one state - return
    #
    if(length(state)==1)
      return(c(start,end,state))
    #
    # keep removing the smallest state until none are less than n2 in length
    #
    m=cbind(start,end,state)
    while(sum((m[,2]-m[,1])<n2)>0)
    {
      q=m[,2]-m[,1]
      l=which.min(q) # remove this sample
      # if there are only two entries, remove smallest
      if(nrow(m)==2)
      {
        mX=m
        m=matrix(m[-l,],ncol=3)
        if(l==2) m[1,2]=mX[2,2]
        if(l==1) m[1,1]=mX[1,1]        
      } else {
        if(l==1)
        {
          m=m[-1,]
          m[1,1]=1
        }
        if(l==nrow(m))
        {
          a=m[nrow(m),2]
          m=m[-nrow(m),]
          m[nrow(m),2]=a
        }
        if(l>1 & l<nrow(m))
        {
          # remove two entries here
          mX=m
          m=matrix(m[-c(l,l+1),],ncol=3)
          m[l-1,2]=mX[l+1,2]
        }
      }
    }
    return(m)
}








###
### in first round, generate a set of positions to check
### in second round, use those positions to see if a break should be made
###
getBetterSwitchesRound1=function(iteration,tempdir,nGen,expRate,sigmaFuture,T,maxRate,L)
{
  ###
  ### calculate breaks using deviation from expected sigma
  ###
  # assuming 0.5 cM/Mb average rate
  dl=diff(L)
  x2=exp(-nGen*expRate/100/1000000/(expRate/maxRate)*dl) # upper
  x3=sigmaFuture
  z1=diff(L) # T-1 length
  #w=1:length(y) #
  a=cbind(1:(T-1),2:T)
  # base pair vs uniform rate is the same... maybe SNP differences
  ### y<0 implies increased rate versus expectation
  #which=y<0
  nAway=50 # how many SNPs apart
  #
  which=x2==x3 # when it equals the max rate
  if(sum(which)==0)
  {
    print(paste("shuffle haplotypes iteration ",iteration," not proceeding - nothing at max recomb rate",sep=""))
  } else {
    b=cbind(a[which,],z1[which]) # col2 is physical distance
    c=dim(b)[1]
    b=cbind(b,c(0,b[-1,1]-b[-c,1]),c(b[-1,1]-b[-c,1],0))
    b=cbind(b,L[b[,1]],L[b[,2]])
    # cluster into groups if >100 SNPs away
    b=cbind(b,1+cumsum(b[,4]>nAway))
    #b2=cbind(b,x2[which]/x1[which],x2[which]/x3[which])
    #b2[100+1:100,c(1,2,3,4,8,9,10)]
    # take the middle positions as where the break occurs
    #
    ### if there are no switches - break here
    #
    if(length(b)>0)
    {
      if(sum(b[,4]>nAway)>0)
      {
        switchPresent=array(0,c(sum(b[,4]>nAway)))
        z=1:dim(b)[1]
        switchStart=switchPresent
        switchEnd=switchPresent
        for(i in 1:length(switchPresent))
        {
          a=z[b[,8]==i]
          switchPresent[i]=b[round(median(a)),1]
          switchStart[i]=b[min(a),1]
          switchEnd[i]=b[max(a),2]
        }
        ### require to be at least 10 SNPs apart
        # also require to be at least 100 away from start, end of chromosome
        which=c(TRUE,diff(switchPresent)>nAway)
        which=which & switchStart>=nAway & switchEnd<=(T-nAway)
        switchPresent=switchPresent[which]
        switchStart=switchStart[which]
        switchEnd=switchEnd[which]
        ### done for now!
        print(paste("iteration - ",iteration," - eligible - ",dim(b)[1]," - n switches - ",length(switchPresent),sep=""))
        # so switchPresent is the rounded median of left points of transitions
        ###
        ### now, for a given set of breaks, calculate what moves where
        ###
        # now - switchPresent is where change is occuring (in index location, not physical)
        # we want two points, one of the left, the other on the right, from which to make the comparison
        breaks1=switchStart-25
        breaks2=switchEnd+25
        #breaks=switchPresent+
        #       round((  x[-length(x)]-x[-1])/2)
        nbreaks=length(breaks1)
        save(switchPresent,breaks1,breaks2,nbreaks,file=paste(tempdir,"switching.iteration",iteration,".RData",sep=""))
        return(1)
      }
    }
  }
  return(0) # nothing to switch
}





getBetterSwitchesSimple=function(fromMat,nbreaks,K,T,eHapsFuture,alphaMatFuture,iteration)  
{
  #which.min(sapply(1:(nbreaks-1),function(i) sum(diag(fromMat[i,,]))))
  # greedy version - choose, in order, one which means fewest moves?
  switchOrder=t(sapply(1:nbreaks,fromMat=fromMat,K=K,function(ib,fromMat,K) {
    fromMatL=fromMat[ib,,]
    # columns are to
    # rows are from
    # ie entry fromMat[10,2,3] is from breaks[10] hap 2to breaks[10+1] hap 3
    #if(ib==10) fromMat[,c(1,2,3)]=
    orderL=array(0,K)
    fromL=1:K
    toL=1:K
    fromR=array(0,K)
    toR=array(0,K)
    for(j in 1:(K-1))
    {
      # choose one which would give best diag
      f=(1:ncol(fromMatL))[rowSums(fromMatL==max(fromMatL))==1]
      t=(1:ncol(fromMatL))[colSums(fromMatL==max(fromMatL))==1]
      fromR[j]=fromL[f]
      toR[j]=toL[t]
      # now - remove from consideration
      fromMatL=fromMatL[-f,-t]
      fromL=fromL[-f]
      toL=toL[-t]
    }
    fromR[K]=fromL
    toR[K]=toL
    # re-order by fromR
    return(toR[order(fromR)])
  }))
  whichIsBest=as.integer(apply(switchOrder,1,function(x) sum(x==(1:K)))==K)
  #
  #
  # determine the order of subsequent regions
  #
  #
  #K=4
  #switchOrder=rbind(c(1,2,3,4),c(2,3,1,4),c(1,2,3,4),c(3,4,1,2))
  #T=500
  breaks=100*(0:(-1+floor(T/100))) + 51  
  nbreaks=length(breaks)-1
  # get the start, end location of the switches
  rStart=100*(0:(-1+floor(T/100)))+1
  rEnd=100*(0:(-1+floor(T/100)))+100
  rEnd[length(rEnd)]=T
  # now - do the unravelling
  tempMat=array(0,c(nbreaks+1,K))
  tempMat[1,]=1:K
  currentState=1:K
  nextState=1:K
  for(iBreak in 1:nbreaks)
  {
    # how this works - first, start in your current state
    # next, choose new state from switchOrder. copy that state, then remember where to go next
    currentState=nextState
    for(k in 1:K)
    {
      # first - the current value is wherever you are
      tempMat[iBreak+1,k]=switchOrder[iBreak,currentState[k]]
      nextState[k]=tempMat[iBreak+1,k]
      # now - record the number for where to start next time
    }
  }
  #
  #
  # do the shuffling
  #
  #
  print(paste("Shuffle haplotypes - Iteration ",iteration," - change ",sum(whichIsBest!=1)," out of ",nbreaks," N=100 SNP intervals",sep="")) # print nuber which changed
  switchCur=1:K
  for(iBreak in 1:(nbreaks+1))
  {
    permL=tempMat[iBreak,]
    eHapsFuture[rStart[iBreak]:rEnd[iBreak],] =eHapsFuture[rStart[iBreak]:rEnd[iBreak],permL]
  }
  for(iBreak in 1:nbreaks)
  {
    if(whichIsBest[iBreak]==0)
    {
      ### add in some noise around break
      ### add in randomness around where I decide to switch
      nSNPL=20
      y=(rEnd[iBreak]-nSNPL):(rEnd[iBreak]+nSNPL)
      eHapsFuture[y,]=runif((2*nSNPL+1)*K)
      alphaMatFuture[y,]=matrix(1/K,ncol=K,nrow=2*nSNPL+1)
    } # end of check on whether we want to make switches
  } # end of injecting randomness back in
  return(list(eHapsFuture=eHapsFuture,alphaMatFuture=alphaMatFuture,switchOrder=switchOrder,whichIsBest=whichIsBest,tempMat=tempMat))
}  




getBetterSwitchesRound2=function(fromMat,K,nGen,expRate,L,eHapsFuture,sigmaFuture,alphaMatFuture,iteration,tempdir)  
{
  # switchPresent,breaks1,breaks2,nbreaks,
  load(file=paste(tempdir,"switching.iteration",iteration-1,".RData",sep=""))
  # greedy version - choose, in order, one which means fewest moves?
  switchOrder=t(sapply(1:nbreaks,fromMat=fromMat,K=K,function(ib,fromMat,K) {
    fromMatL=fromMat[ib,,]
    orderL=array(0,K)
    fromL=1:K
    toL=1:K
    fromR=array(0,K)
    toR=array(0,K)
    for(j in 1:(K-1))
    {
      # choose one which would give best diag
      f=(1:ncol(fromMatL))[rowSums(fromMatL==max(fromMatL))==1]
      t=(1:ncol(fromMatL))[colSums(fromMatL==max(fromMatL))==1]
      fromR[j]=fromL[f]
      toR[j]=toL[t]
      # now - remove from consideration
      fromMatL=fromMatL[-f,-t]
      fromL=fromL[-f]
      toL=toL[-t]
    }
    fromR[K]=fromL
    toR[K]=toL
    # re-order by fromR
    return(toR[order(fromR)])
  })) # end of switch order
  #
  #
  #
  whichIsBest=as.integer(apply(switchOrder,1,function(x) sum(x==(1:K)))==K)
  switchCur=1:K
  switchPresentAll=c(1,switchPresent,T)
  tempMat=array(0,c(nbreaks+1,K))
  tempMat[1,]=1:K
  print(paste("shuffle haplotypes - iteration - ",iteration," - change - ",sum(whichIsBest!=1),sep="")) # print nuber which changed
  for(iBreak in 1:nbreaks)
  {
    # for example, when iBreak is 1
    # am resolving from 1 to 2
    # use fromMat[1,,]
    #### OLD VERSIOn
    #permL=permM[whichIsBest[iBreak],][switchCur]
    ### NEW VERSION
    permL=switchOrder[iBreak,]
    tempMat[iBreak+1,]=permL
    #test[switchPresentAll[iBreak]:switchPresentAll[iBreak+1]]=iBreak
    eHapsFuture[switchPresentAll[iBreak+1]:switchPresentAll[iBreak+2],] =
    eHapsFuture[switchPresentAll[iBreak+1]:switchPresentAll[iBreak+2],permL]
    switchCur=permL
  }
  ### add in randomness around where I decide to switch
  for(iBreak in (1:(nbreaks))[whichIsBest>1])
  {
    a=matrix(runif((breaks2[iBreak]-breaks1[iBreak]+1)*K),ncol=K)
    a=a/rowSums(a)
    eHapsFuture[breaks1[iBreak]:breaks2[iBreak],]=a
    # reset sigmaList as well
    # note that this is kind of off by 1 - uses right part of switch
    sigmaFuture[breaks1[iBreak]:breaks2[iBreak]]=exp(-nGen*expRate/100/1000000*diff(L)) # interior is morgans between
    alphaMatFuture[breaks1[iBreak]:breaks2[iBreak],]=
      matrix(1/K,nrow=length(breaks1[iBreak]:breaks2[iBreak]),ncol=K)
  }
  return(list(eHapsFuture=eHapsFuture,sigmaFuture=sigmaFuture,alphaMatFuture=alphaMatFuture))
}  










# derived from "A Note on Exact Tests of Hardy-Weinberg Equilibrium"
calculate_hwe_p <- function(x) {
  #x <- c(100,50,10)

  if (x[3]>x[1])
    x <- x[c(3,2,1)]
  
  nAA <- x[1] # individuals
  nAB <- x[2] # individuals
  nBB <- x[3] # individuals
  nA <- nAA * 2 + nAB # number of copies of A allele
  nB <- nBB * 2 + nAB # number of copies of B allele
  n <- nAA + nAB + nBB # number of (diploid) individuals
  
  min_het <- 0
  max_het <- nAB + 2 * min(nAA, nBB)

  # maximum-ish value
  mid <- floor((nA * nB) / ((nA + nB)))
  if (is.na(mid))
    mid <- 0
  # make odd if it needs to be
  if ((nA%%2) != (mid%%2))
    mid <- mid + 1
  
  # determine a mid point
  probs <- array(0, max_het + 1) # note - this is 0-based
  probs[mid + 1] <- 1
  
  # use recurrence relation - going down
  n_het <- mid
  n_hom_alt <- (nBB * 2 + nAB - n_het) / 2
  n_hom_ref <- (n - n_het - n_hom_alt)
  if ((mid - 2) >= min_het) {
    for(het in seq(mid - 2, min_het, -2)) {
      probs[het + 1] <- probs[het + 3] *
        n_het * (n_het-1) / (4 * (n_hom_ref + 1) * (n_hom_alt + 1))
      n_het <- n_het - 2
      n_hom_ref <- n_hom_ref + 1
      n_hom_alt <- n_hom_alt + 1
    }
  }
    
  # use recurrence relationship - going up
  n_het <- mid
  n_hom_alt <- (nBB * 2 + nAB - n_het) / 2
  n_hom_ref <- (n - n_het - n_hom_alt)
  if ((mid + 2) <= max_het) {  
    for(het in seq(mid + 2, max_het, 2)) {
      probs[het + 1] <- probs[het - 1] *
        (4 * n_hom_ref * n_hom_alt) / ( (n_het + 2) * (n_het + 1))
      n_het <- n_het + 2
      n_hom_ref <- n_hom_ref - 1
      n_hom_alt <- n_hom_alt - 1
    }
  }
    
  all_probs <- probs / sum(probs)
  p <- min(1, sum(all_probs[all_probs <= all_probs[nAB + 1]]))
  return(p)
}




generate_hwe_on_counts <- function(
  hweCount,
  T,
  nCores
) {
  # few SNPs - make at least 5 per entry
  x=seq(1,T,by=ceiling(T/nCores))    
  if(ceiling(T/nCores)<10)  x=seq(1,T,by=ceiling(T/10))
  x2=c(x[-1]-1,T)
  x3=as.list(1:nCores)
  for(i in 1:length(x2))
    x3[[i]]= c(x[i],x2[i])
  x3=x3[unlist(lapply(x3,length))>1] # make sure each one has at least two to load  
  out <- mclapply(x3,hweCount=hweCount,function(i, hweCount) {
    return(apply(hweCount[i[1]:i[2],],1,calculate_hwe_p))
  })
  out <- unlist(out)
  return(out)
}



# function to get phred scaled likelihood and genotype
# from input RData file
get_pl_and_rd_from_tempdir <- function(iSample,tempdir,regionName,T)
{
  # load sample reads
  load(file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))  
  # get genotype likelihoods
  x=unlist(sapply(sampleReads,function(x) x[[4]]))+1
  bq=unlist(sapply(sampleReads,function(x) x[[3]]))
  ### MODIFYING
  # convert to original BQs
    #bq=(1-y)/(1-2/3*y)
    #bq2=y/(1/3+2/3*y)
    #bq[bq2<0.5]=bq2[bq2<0.5]
    #bq=-10*log10(bq)
    #s=as.integer(y>0.5) # this was 1=alt, 0 = ref
    s=as.integer(bq>0)
    bq=abs(bq)
    e=10^(-bq/10)
    p=cbind(1-e,e*(1/3))
    pA=p[cbind(1:nrow(p),s+1)]
    pB=p[cbind(1:nrow(p),2-s)]    
    # hmm, just do for loop? not ideal, but OK?
    g=array(1,c(T,3))
    rd=array(0,c(T,2))
    for(i in 1:length(x)) {
      a=x[i]
      #b=y[i]
      b=pA[i]/2
      c=pB[i]/2    
      g[a,]=g[a,] * c(b+b, b+c, c+c) # new
      z <- which.max(c(pA[i], pB[i]))
      rd[a,z] = rd[a,z] + 1
    }
    g=g/rowSums(g) # divide by max
    a=g[,1]
    w=a<g[,2]; a[w]=g[w,2]
    w=a<g[,3]; a[w]=g[w,3]    
    g=g/a
    # 0/0:2,0:2:3:0,3,45 
    # turn into PLs
    # ex: 1/1:0,1:1:3:32,3,0    
    pl=round(10 * -log10(g) )
    return(list(pl = pl, rd = rd))
}
  

# if want to output genotype probabilities in VCF format
outputInputInVCFFunction <- function(
  outputdir,
  pos,
  T,
  tempdir,
  N,
  nCores,
  regionName,
  environment,
  sampleNames,
  outputBlockSize
) {
  #
  # need to loop over samples, very similar to normal output
  # write out chunks
  # except this time, want genotype, allelic depth, pl
  # gt,ad,pl
  #
  #
  #
  #  
  # 1 - build the data
  #
  #
  #
  #
  x3 <- getSampleRange(N, nCores)
  f <- function(sampleRange, N, outputBlockSize, tempdir, regionName, T) {
    i_core <- match(sampleRange[1], sapply(x3, function(x) x[[1]]))
    list_of_vcf_columns_to_out <- as.list(1:N)
    outputBlockRange <- getOutputBlockRange(sampleRange,outputBlockSize )
    for(iSample in sampleRange[1]:sampleRange[2]) {
      o <- get_pl_and_rd_from_tempdir(iSample,tempdir,regionName,T)
      pl <- o$pl
      rd <- o$rd
      list_of_vcf_columns_to_out[[iSample]] <- make_column_of_vcf_from_pl_and_rd(pl, rd)
      # if appropriate - write to disk!
      iBlock <- match(iSample, outputBlockRange)      
      if (iBlock > 1 & is.na(iBlock)==FALSE) {
        iSample_list <- (outputBlockRange[iBlock-1]+1):outputBlockRange[iBlock]            
        write_block_of_vcf(i_core, iBlock, list_of_vcf_columns_to_out, iSample_list = iSample_list, outputdir, regionName)
        list_of_vcf_columns_to_out <- as.list(1:N)
      }
    }
  }
  #
  # loop over samples - write to disk!
  #
  if(environment=="server")
  {
    out2=mclapply(x3,mc.cores=nCores,FUN=f, N = N, outputBlockSize = outputBlockSize, tempdir = tempdir, regionName = regionName, T = T)
  }
  if(environment=="cluster")
  {
    cl = makeCluster(nCores, type = "FORK")
    out2 = parLapply(cl, x3, fun=f,N = N, outputBlockSize = outputBlockSize, tempdir = tempdir, regionName = regionName, T = T)
    stopCluster(cl)
  }
  #
  #
  #
  #
  # step 2 - build the VCF itself
  #
  #
  #
  #
  # build header
  # build left side
  # stitch everything together
  print(paste0("Build vcf from input - ",date()))
  output_vcf_header <- paste0(outputdir,"stitch.input.",regionName,".header.vcf")
  output_vcf_left <- paste0(outputdir,"stitch.input.",regionName,".left.vcf.gz")
  output_vcf <- paste0(outputdir,"stitch.input.",regionName,".vcf.gz")
  # 
  # do header lines
  #
  header <- paste0(
    '##fileformat=VCFv4.0\n',
    '##FORMAT=<ID=AD,Number=.,Type=Integer,Description="Allelic depths for the ref and alt alleles in the order listed">\n',
    '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n',
    '##FORMAT=<ID=PL,Number=G,Type=Integer,Description="Normalized, Phred-scaled likelihoods for genotypes as defined in the VCF specification">\n'
  )
  header2 <- paste("#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT", paste(sampleNames, collapse = "\t", sep="\t"), sep="\t")
  #
  # add output
  #
  cat(header, header2, "\n", sep="", file = output_vcf_header)
  #
  # build first part of file
  #
  options(scipen=6)
  INFO <- "."
  write.table(
    matrix(paste(pos[,1], pos[,2], ".", pos[,3], pos[,4], ".", "PASS", INFO , "GT:AD:PL", sep="\t"), ncol=1),
    file = gzfile(output_vcf_left),
    row.names = FALSE,
    col.names = FALSE,
    sep = "",
    quote = FALSE
  )
  #
  # get number of cores and blocks
  #
  x3 <- getSampleRange(N, nCores)  
  files_to_paste <- lapply(1:nCores, function(i_core) {
    nBlocks <- length(getOutputBlockRange(
      sampleRange = x3[[i_core]],
      outputBlockSize
    ))
    return(paste0('<(zless ',outputdir, "vcf.piece.",i_core,".",2:nBlocks,".",regionName,".txt.gz", ' | cat ) '))
})
  files_to_paste <- paste(unlist(files_to_paste), collapse = " ")
  # 
  # merge together
  #
  command <- paste0(
    'bash -c "',          
    'paste -d ',shQuote("\t"),' ',
    '<(zless ',output_vcf_left,' | cat ) ',
    files_to_paste, " | ",
    " cat ",output_vcf_header," - | gzip -1 > ",
      output_vcf
    ,'"'
  )
  #print(command)
  system(
    command
  )
  system(paste0("rm ",output_vcf_header))
  system(paste0("rm ",output_vcf_left))
  #
  # remove
  #
  x3 <- getSampleRange(N, nCores)  
  files_to_remove <- unlist(lapply(1:nCores, function(i_core) {
    nBlocks <- length(getOutputBlockRange(
      sampleRange = x3[[i_core]],
      outputBlockSize
    ))
    return(paste0(outputdir, "vcf.piece.",i_core,".",2:nBlocks,".",regionName,".txt.gz"))
  }))
  system(paste0("rm ",paste0(files_to_remove, collapse=" ")))
}






### get the r2's for all 4 high coverage mice and the average (for an iteration, ie easy)
getR2DuringEMAlgorithm=function(highCovInLow,gen,K,T,iteration,runVersion,chr,tempdir,nHC,regionName)
{
  out=as.list(1:nHC)
  for(j in 1:nHC)
  {
    load(file=paste(tempdir,"sample.",highCovInLow[j],".input.",regionName,".dosage.RData",sep=""))
    out[[j]]=dosage
  }
  # now, get r2 between generated and observed, store and output
  store=array(0,nHC+1)
  for(isample in 1:nHC)
  {
    m1=cbind(gen[,isample]/2,out[[isample]])
    store[isample]=c(cor(m1[,1],m1[,2],use="complete.obs")^2)
  }
  store[nHC+1]=mean(store[1:nHC])
  return(store)
}





### plot HWE, info, MAF, coloured by discrepency
plotMetricsForPostImputationQC=function(iSample,highCovList,gen,gen_imp,alleleCount,chr,L,estimatedAlleleFrequency,info,outputdir,colour=TRUE,hwe,regionName)
{
  if(is.null(iSample)==FALSE)
  {
    m1A=cbind(gen[,iSample]/2,gen_imp[, iSample])
    m1A[alleleCount[,3]>0.5,]=1-m1A[alleleCount[,3]>0.5,]
    dist=abs(m1A[,1]-m1A[,2])
    dist[dist>0.1]=0.1
  } else {
    dist=rep(0,T)
  }
  colfunc <- colorRampPalette(c("blue", "red"))
  col=colfunc(11)[round(dist,2)*100+1]
  ### plot
  jpeg(paste(outputdir,"plots/metricsForPostImputationQC.",regionName,".sample",iSample,".jpg",sep=""),height=1000,width=3000,qual=100)
  par(mfrow=c(1,3))
  plot(info,log10(hwe) ,cex.lab=3 ,col=col)
  plot(info,estimatedAlleleFrequency,cex.lab=3,col=col)
  plot(estimatedAlleleFrequency,log10(hwe) ,cex.lab=3,col=col )
  dev.off()
  #sum(log10(hwe)>-6)
  ### now plot along the chromosome as well
  jpeg(paste(outputdir,"plots/metricsForPostImputationQCChromosomeWide.",regionName,".sample",iSample,".jpg",sep=""),height=3000,width=3000,qual=100)
  par(mfrow=c(3,1))
  plot(L,log10(hwe) ,cex.lab=3 ,col=col)
  plot(L,info,cex.lab=3,col=col)
  plot(L,estimatedAlleleFrequency,cex.lab=3,col=col )
  dev.off()
  #sum(log10(hwe)>-6)
}

### plot hapSum along genome
plotHapSum=function(outname,L,which,K,hapSum,T,plotExtra=FALSE,switchPresent=NULL,breaks=NULL,plotR=FALSE,simCor=NA,hsd=NA,N)
{
  jpeg(outname,height=(1+as.integer(plotR==TRUE))*2000,width=10000,qual=100)
  if(plotR==TRUE) par(mfrow=c(2,1))
  colStore=rep(c("black","red","green","blue"),ceiling(K/4))
  sum=array(0,c(length(which)))
  xlim=range(L[which])
  ylim=c(0,1)
  plot(x=L[1],y=0,xlim=xlim,ylim=ylim,axes=FALSE)
  x=c(L[which[1]],L[which],L[which[length(which):1]])
  # build sums first
  m=array(0,c(length(which),K+1))
  for(i in 1:K)
    m[,i+1]=m[,i]+hapSum[which,i]/N
  for(i in K:1)
  {
    polygon(x=x,y=c(m[1,i],m[,i+1],m[length(which):1,i]),xlim=xlim,ylim=ylim,
            col=colStore[i])
  }
  if(plotR==TRUE)
  {
    plot(x=L,y=simCor + rnorm(T,sd=1e-2),pch=16,xlim=xlim,ylim=ylim,axes=FALSE,col=rgb(1, 0, 0, 0.33))
    par(new=TRUE)
    plot(x=hsd[,1],y=ylim[1] + diff(ylim)* hsd[,2]/diff(range(hsd[,2])),type="l",xlim=xlim,ylim=ylim,axes=FALSE,lwd=2)    
  }
  if(plotExtra==TRUE)
  {
    for(l in   L[switchPresent])
      abline(v=l,lwd=10)
    for(l in   L[breaks])
      abline(v=l,lwd=10,col="purple")
  }
  dev.off()
}




### plot estimated against actual allele frequency
plotEstimatedAgainstReal=function(outputdir,alleleCount,estimatedAlleleFrequency,which,chr,regionName)
{
  m1=cbind(alleleCount[,3],estimatedAlleleFrequency)
  jpeg(paste(outputdir,"plots/r2.",regionName,".goodonly.jpg",sep=""),height=800,width=800,qual=100)
  m1=m1[which,]
  main=paste("nSnps: ",dim(m1)[1],", r2:",round(cor(m1[,1],m1[,2],use="complete.obs")^2,3),sep="")
  plot(x=m1[,1],y=m1[,2],xlim=c(0,1),ylim=c(0,1),main=main,xlab="Real Allele Frequency",ylab="Estimated allele frequency",pch=20,cex.main=1.5,cex.lab=1.5)
  abline(0,1)
  dev.off()
}



findRecombinedReadsPerSample=function(fbsoL,gamma,eHapsCurrent,K,L,iSample,verbose=FALSE,sampleReads,tempdir,N,regionName)
{
  #
  # written for diploid version currently
  #
  improvement=lapply(1:N,function(x) list())
  # needs a full run
  # only do for some - need at least 3 SNPs to consider
  whichReads=unlist(lapply(sampleReads,function(x) x[[1]]))>2
  h=function(x) {
    y=cbind(matrix(eHapsCurrent[x[[4]]+1,],ncol=K),1/2)
    sapply(1:(K+1),function(k)
      exp(sum(log(x[[3]]*y[,k] + (1-x[[3]])*(1-y[,k])))))
  }
  m=t(sapply(sampleReads[whichReads],h))
  #
  #
  # if any do more poorly than a "random" read, then try splitting
  #
  #
  count=0
  w=(1:length(sampleReads))[whichReads][(apply(m,1,which.max)==(K+1))]
  # can't do first or last read
  w=w[w>1 & w<length(sampleReads)]
  if(length(w)>0)
  {
    for(w1 in w)
    {
      # go forward, backward, 5 reads
      b=c(max(1,w1-10),min(length(sampleReads),w1+10))
      a=c(sampleReads[[b[1]]][[2]]+1,sampleReads[[b[2]]][[2]]+1)
      #a=sampleReads[[w1]][[2]]+c(-1000,1000)
      #a=sampleReads[[w1]][[2]]
      #
      # NOTE - for now - do easy thing - add together
      #
      change=apply(gamma[a,],2,diff)
      to=which.max(change)
      from=which.min(change)          
      # try a break between all SNPs - take the best one
      x=sampleReads[[w1]]
      y=cbind(matrix(eHapsCurrent[x[[4]]+1,],ncol=K))
      bqProbs=convertScaledBQtoProbs(x[[3]])
      g1=bqProbs[,2]*y[,from] + bqProbs[,1]*(1-y[,from])
      g2=bqProbs[,2]*y[,to]   + bqProbs[,1]*(1-y[,to])
      #
      z=sapply(2:(x[[1]]),function(a) {
        x1=(sum(log(g1[1:(a-1)])))
        x2=(sum(log(g2[(a+1):x[[1]]])))
        sum(x1+x2)
      })
      #
      # if better than random - break
      #
      u=log(h(sampleReads[[w1]]))[K+1]
      if(max(z)>u)
      {
        improvement=list(c(max(z),u,length(x[[3]])))
        #
        g=function(w22) {
          r=which.min(abs(L[x[[4]][w22]+1]-mean(L[x[[4]][w22]+1])))
          return(list(length(w22)-1,x[[4]][w22][r],x[[3]][w22],x[[4]][w22])) }
        #
        sampleReads[[w1]]=g(1:which.max(z))
        sampleReads2=as.list(1:length(sampleReads)+1)
        sampleReads2[1:(length(sampleReads))]=sampleReads
        sampleReads2[[length(sampleReads2)]]=g((which.max(z)+1):(x[[1]]+1))
        sampleReads=sampleReads2
        count=count+1
      } # end of check on whether its bigger
    } # end of loop on reads
  } # end of check on whether there are reads to split
  # now that we've updated reads, re-order them
  sampleReads=sampleReads[order(unlist(lapply(sampleReads,function(x) x[[2]])))]
  #
  # save output back onto file in tempdir
  #
  save(sampleReads,file=paste(tempdir,"sample.",iSample,".input.",regionName,".RData",sep=""))
  if(verbose==TRUE) print(paste("sample ",iSample," readsSplit ",count," readsTotal ",length(sampleReads),sep=""))  
  return(list(readsSplit=count,readsTotal=length(sampleReads)))
}




pseudoHaploidUpdate=function(pRgivenH1,pRgivenH2,fbsoL,pseudoHaploidModel,srp,delta)
{
      #
      # once both haplotypes are available - update omega immediately
      #
      # for each read, get new phasing
      # srp comes in with readProbs originally - doesnt change
      # 4 things here
      # 1 - pRgivenH1 - P(R | h1)
      # 2 - pRgivenH2 - P(R | h2)
      # note that
      # 3 - pRandH1 - P(R and H1) = 1/2 P(R| H1)
      # 4 - pRandH2 - P(R and H2) = 1/2 P(R| H2)
      ###
      ### MODEL 1 - divide one by other (y1/(y1+y2)) to get readProbs
      ###
      # readProbs is y1/(y1+y2) from above
      # then probabilities using
      #    eMatHap(iRead,k) = eMatHap(iRead,k) * readProbs(iRead) + (1/Kd)  * (1-readProbs(iRead));
      # then updating using multiply both by readProbs
      # y1=(fbsoL[[1]]$gamma[srp+1,] * fbsoL[[1]]$eMatHap)
      # y2=(fbsoL[[2]]$gamma[srp+1,] * fbsoL[[2]]$eMatHap)
      # readProbs=y1/(y1+y2)
      ###
      ### MODEL 3, 7 - try to predict which haplotype it came from
      ###
      if(pseudoHaploidModel==7 | pseudoHaploidModel==9)
      {
        x=pRgivenH1/(pRgivenH1+pRgivenH2)
        # get eMatHap without the second part
        y1=(fbsoL[[1]]$eMatHap - (1-x)*pRgivenH2)/x
        y2=(fbsoL[[2]]$eMatHap - (x)*pRgivenH1)/(1-x)
        pRgivenH1=rowSums(fbsoL[[1]]$gamma[srp+1,] * y1)
        pRgivenH2=rowSums(fbsoL[[2]]$gamma[srp+1,] * y2)
        # bound above below
        pRgivenH1[pRgivenH1<0.001]=0.001
        pRgivenH2[pRgivenH2<0.001]=0.001
        pRgivenH1[pRgivenH1>0.999]=0.999
        pRgivenH2[pRgivenH2>0.999]=0.999
      }
      ###
      ### MODEL 8
      ###
      if(pseudoHaploidModel==8 | pseudoHaploidModel==10)
      {
        x=pRgivenH1/(pRgivenH1+pRgivenH2)
        # get eMatHap without the second part
        y1=fbsoL[[1]]$eMatHapOri # shouldn't change 1 or 2
        pRgivenH1=rowSums(fbsoL[[1]]$gamma[srp+1,] * y1)
        pRgivenH2=rowSums(fbsoL[[2]]$gamma[srp+1,] * y1)
        # bound above below
        pRgivenH1[pRgivenH1<0.001]=0.001
        pRgivenH2[pRgivenH2<0.001]=0.001
        pRgivenH1[pRgivenH1>0.999]=0.999
        pRgivenH2[pRgivenH2>0.999]=0.999
      }
      ###
      ### MODEL 4 - 
      ###
      if(pseudoHaploidModel==4)
      {
        # model 3 - 
        #x=0.5;
        #for(k=0; k<=K-1; k++)
        #  eMatHap(iRead,k) = x * eMatHap(iRead,k) + (1-x) * pRgivenH2(iRead);
        # update using proportions 
        x=0.5;      
        # get eMatHap without the second part
        y1=(fbsoL[[1]]$eMatHap - (1-x)*pRgivenH2)/x
        y2=(fbsoL[[2]]$eMatHap - (x)*pRgivenH1)/(1-x)
        pRgivenH1=rowSums(fbsoL[[1]]$gamma[srp+1,] * y1)
        pRgivenH2=rowSums(fbsoL[[2]]$gamma[srp+1,] * y2)
        # bound above below
        pRgivenH1[pRgivenH1<0.001]=0.001
        pRgivenH2[pRgivenH2<0.001]=0.001
        pRgivenH1[pRgivenH1>0.999]=0.999
        pRgivenH2[pRgivenH2>0.999]=0.999
      }
      ###
      ### MODEL 5 - complete proportioning
      ###
      if(pseudoHaploidModel==5)
      {
        # update using proportions
        x=pRgivenH1/(pRgivenH1+pRgivenH2)        
        # get eMatHap without the second part
        y1=(fbsoL[[1]]$eMatHap - (1-x)*pRgivenH2)/x
        y2=(fbsoL[[2]]$eMatHap - (x)*pRgivenH1)/(1-x)
        # might have NA's - if so, replace one with other
        y=array(0,dim(y1))
        w=is.na(y1[,1])==FALSE; y[w,]=y1[w,]
        w=is.na(y2[,1])==FALSE; y[w,]=y2[w,]        
        a1=rowSums(fbsoL[[1]]$gamma[srp+1,] * y)
        a2=rowSums(fbsoL[[2]]$gamma[srp+1,] * y)
        pRgivenH1=as.integer(runif(length(srp))<a1/(a1+a2))
        pRgivenH2=1-pRgivenH1
      }
      ###
      ### MODEL 6 - fancier method
      ###
      if(pseudoHaploidModel==6)
      {
        a=(rowSums(fbsoL[[1]]$gammaUpdateS1) + fbsoL[[2]]$gammaUpdateS2)/2
        b=(rowSums(fbsoL[[2]]$gammaUpdateS1) + fbsoL[[1]]$gammaUpdateS2)/2        
        delta=a/(a+b)
        x=pRgivenH1/(pRgivenH1+pRgivenH2)
        # get eMatHap without the second part
        y1=(fbsoL[[1]]$eMatHap - (1-x)*pRgivenH2)/x
        y2=(fbsoL[[2]]$eMatHap - (x)*pRgivenH1)/(1-x)
        pRgivenH1=rowSums(fbsoL[[1]]$gamma[srp+1,] * y1)
        pRgivenH2=rowSums(fbsoL[[2]]$gamma[srp+1,] * y2)
        # bound above below
        pRgivenH1[pRgivenH1<0.001]=0.001
        pRgivenH2[pRgivenH2<0.001]=0.001
        pRgivenH1[pRgivenH1>0.999]=0.999
        pRgivenH2[pRgivenH2>0.999]=0.999
        #pRandH1=pRgivenH1/(pRgivenH1+pRgivenH2)
        #pRandH2=pRgivenH2/(pRgivenH1+pRgivenH2)
      }
  return(list(pRgivenH1=pRgivenH1,pRgivenH2=pRgivenH2,delta=delta))
}


# on final iteration, get imputed dosages
get_high_coverage_final_imputed_genotypes <- function(
  highCovInLow,
  nHC,
  tempdir,
  regionName,
  T
) {
  out <- array(NA, c(T, nHC))
  for(j in 1:nHC)
  {
    load(file=paste(tempdir,"sample.",highCovInLow[j],".input.",regionName,".dosage.RData",sep=""))
    out[, j] <- dosage
  }
  return(out)
}


get_max_gen_rapid <- function(x) {
  # assume matrix 3 columns >1 row
  z <- rep(1, nrow(x))
  y <- x[,1]
  for(i in 2:3) {
    w <- x[,i]>y
    z[w] <- i
    y[w] <- x[w,i]
  }
  return(cbind(1:nrow(x), z))
}

# write out a single samples worth of VCF entry
make_column_of_vcf <- function(
  gp
) {
  # write out genotype, genotype likelihood, and dosage
  #GT:GL:DS    
  ##FORMAT=<ID=GT:,Number=1,Type=String,Description="Best Guessed Genotype with posterior probability threshold of 0.9">
  ##FORMAT=<ID=GL,Number=3,Type=Float,Description="Posterior probability of 0/0, 0/1, and 1/1">
  ##FORMAT=<ID=DS,Number=1,Type=Float,Description="Dosage">
  # 1/1:0,0.054,0.946:1.946
  # add one samples worth of info to a VCF
  z <- get_max_gen_rapid(gp)
  gt <- c("0/0","0/1","1/1")[z[,2]]
  gt[gp[z] < 0.9] <- "./."
  precision <- 3
  str <- paste0(
    gt,":",
      round(gp[,1], precision), ",",
      round(gp[,2], precision), ",",
      round(gp[,3], precision), ":",
    round(gp[,2] + 2 * gp[,3], precision)
  )
  return(str)
}


# write out a single samples worth of VCF entry
make_column_of_vcf_from_pl_and_rd <- function(
  pl,
  rd
) {
  # write out genotype, genotype likelihood, and dosage
  # gt, ad, pl
  # GT:AD:PL
  #chr19   3126522 .       G       A       7661.22 .       AC=464;AF=0.935;AN=496;BaseQRankSum=4.232;DP=333;Dels=0.00;FS=2.773;GC=39.90;HRun=0;HaplotypeScore=0.0738;InbreedingCoeff=-0.0083;MLEAC=465;MLEAF=0.938;MQ=42.81;MQ0=23;MQRankSum=-0.576;QD=28.69;ReadPosRankSum=-0.673      GT:AD:DP:GQ:PL       ./.     ./.     ./.     1/1:0,1:1:3:32,3,0      ./.     ./.
  str <- paste0(
    "./.",":",
    rd[,1], ",",
    rd[,2], ":" ,
    pl[,1], ",",
    pl[,2], "," ,
    pl[,3]
  )
  str[(rd[,1] + rd[,2])==0] = "./."
  return(str)
}


# write a block of samples to disk
# don't write header or row information 
write_block_of_vcf <- function(
  i_core,
  iBlock,
  list_of_vcf_columns_to_out,
  iSample_list,
  outputdir,
  regionName
) {
  print(paste("Write block of VCF for samples:",iSample_list[1],"-",iSample_list[length(iSample_list)], " - ", date(), sep=""))
  m <- matrix(sapply(list_of_vcf_columns_to_out[iSample_list],I), ncol = length(iSample_list))    
  write.table(
    m,
    file = gzfile(paste0(outputdir, "vcf.piece.",i_core,".",iBlock,".",regionName,".txt.gz")),
    row.names = FALSE,
    col.names = FALSE,
    sep = "\t",
    quote = FALSE
  )
  return(NULL)
}



# build the header
# build the left part of the VCF
# paste together the interim VCF files
# gzip the whole thing together
write_vcf_after_EM <- function(
  outputdir,
  regionName,
  sampleNames,
  tempdir,
  nCores,
  info,
  hwe,
  estimatedAlleleFrequency,
  pos,
  N,
  outputBlockSize
) {
  #
  print(paste0("Build final vcf - ",date()))
  output_vcf_header <- paste0(outputdir,"stitch.",regionName,".header.vcf")
  output_vcf_left <- paste0(outputdir,"stitch.",regionName,".left.vcf.gz")
  output_vcf <- paste0(outputdir,"stitch.",regionName,".vcf.gz")
  # 
  # do header lines
  #
  header <- paste0(
    '##fileformat=VCFv4.0\n',
    '##INFO=<ID=INFO_SCORE,Number=.,Type=Float,Description="Info score">\n',
    '##INFO=<ID=EAF,Number=.,Type=Float,Description="Estimated allele frequency">\n',
    '##INFO=<ID=HWE,Number=.,Type=Float,Description="Hardy-Weinberg p-value">\n',
    '##FORMAT=<ID=GT,Number=1,Type=String,Description="Best Guessed Genotype with posterior probability threshold of 0.9">\n',
    '##FORMAT=<ID=GP,Number=3,Type=Float,Description="Posterior genotype probability of 0/0, 0/1, and 1/1">\n',
    '##FORMAT=<ID=DS,Number=1,Type=Float,Description="Dosage">\n'
  )
  header2 <- paste("#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT", paste(sampleNames, collapse = "\t", sep="\t"), sep="\t")
  #
  # add output
  #
  cat(header, header2, "\n", sep="", file = output_vcf_header)
  #
  # build first part of file
  #
  options(scipen=6)
  INFO <- paste0(
    "EAF=", round(estimatedAlleleFrequency, 5), ";",
    "INFO_SCORE=", round(info, 5), ";",
    "HWE=", hwe
  )
  write.table(
    matrix(paste(pos[,1], pos[,2], ".", pos[,3], pos[,4], ".", "PASS", INFO, "GT:GP:DS", sep="\t"), ncol=1),
    file = gzfile(output_vcf_left),
    row.names = FALSE,
    col.names = FALSE,
    sep = "",
    quote = FALSE
  )
  #
  # get number of cores and blocks
  x3 <- getSampleRange(N, nCores)  
  files_to_paste <- lapply(1:nCores, function(i_core) {
    nBlocks <- length(getOutputBlockRange(
      sampleRange = x3[[i_core]],
      outputBlockSize
    ))
    return(paste0('<(zless ',outputdir, "vcf.piece.",i_core,".",2:nBlocks,".",regionName,".txt.gz", ' | cat ) '))
})
  files_to_paste <- paste(unlist(files_to_paste), collapse = " ")
  # 
  # merge together
  #
  command <- paste0(
    'bash -c "',          
    'paste -d ',shQuote("\t"),' ',
    '<(zless ',output_vcf_left,' | cat ) ',
    files_to_paste, " | ",
    " cat ",output_vcf_header," - | gzip -1 > ",
      output_vcf
    ,'"'
  )
  #print(command)
  system(
    command
  )
  system(paste0("rm ",output_vcf_header))
  system(paste0("rm ",output_vcf_left))
  #
  # remove
  #
  x3 <- getSampleRange(N, nCores)  
  files_to_remove <- unlist(lapply(1:nCores, function(i_core) {
    nBlocks <- length(getOutputBlockRange(
      sampleRange = x3[[i_core]],
      outputBlockSize
    ))
    return(paste0(outputdir, "vcf.piece.",i_core,".",2:nBlocks,".",regionName,".txt.gz"))
  }))
  system(paste0("rm ",paste0(files_to_remove, collapse=" ")))
  return(NULL)
}
